#!/usr/bin/env python3
#
# Copyright 2008, 2009 Hannes Hochreiner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.
#
"""
Jessyink Set Master Slide
"""

import inkex

from jessyink_install import JessyInkMixin, _


class MasterSlide(JessyInkMixin, inkex.EffectExtension):
    """Effect Extension for master slide"""

    def add_arguments(self, pars):
        self.arg_parser.add_argument("--tab")
        self.arg_parser.add_argument("--layerName", default="")

    def effect(self):
        self.is_installed()
        # Remove old master slide property
        for node in self.svg.xpath("//*[@jessyink:masterSlide='masterSlide']"):
            node.set("jessyink:masterSlide", None)

        # Set new master slide.
        if self.options.layerName != "":
            nodes = self.svg.xpath(
                f"//*[@inkscape:groupmode='layer' "
                f"and @inkscape:label='{self.options.layerName}']"
            )
            if not nodes:
                inkex.errormsg(
                    _("Layer not found. Removed current master slide selection.\n")
                )
            elif len(nodes) > 1:
                inkex.errormsg(
                    _(
                        "More than one layer with this name found. "
                        "Removed current master slide selection.\n"
                    )
                )
            else:
                nodes[0].set("jessyink:masterSlide", "masterSlide")


if __name__ == "__main__":
    MasterSlide().run()



"""
VI text tools

Decodes Linearity Curve styledTexts and turn them into simpler format.
"""

from typing import Dict, List

from inkex.utils import debug


def decode_new_text(styled_text: Dict) -> List[Dict]:
    """Decodes upperBound system used by Newer text format."""
    # scanning upperBounds in styledText
    upper_bounds: List[int] = []
    for key in styled_text.keys():
        # TODO skipped strokeStyle
        # debug(f"KeyName:{key}")
        if (
            isinstance(styled_text[key], dict)
            and styled_text[key].get("values") is not None
        ):
            values = styled_text[key]["values"]
            for child_value in values:
                upper_bound = child_value["upperBound"]
                if upper_bound not in upper_bounds:
                    upper_bounds.append(upper_bound)
    upper_bounds.sort()

    # unifying each styles in styledText
    styles: List[Dict] = [{} for _ in range(len(upper_bounds))]
    for key in styled_text.keys():
        if (
            isinstance(styled_text[key], dict)
            and styled_text[key].get("values") is not None
        ):
            values = styled_text[key]["values"]
            for child_value in values:
                upper_bound = child_value["upperBound"]
                index = upper_bounds.index(upper_bound)
                styles[index][key] = child_value["value"]
                if index == 0:
                    styles[index]["length"] = upper_bound
                else:
                    styles[index]["length"] = upper_bound - upper_bounds[index - 1]

    # Propagate last known values forward
    last_known_values = {}
    for i in range(len(styles) - 1, -1, -1):
        for key in styled_text.keys():
            if key in styles[i]:
                last_known_values[key] = styles[i][key]
            elif key in last_known_values:
                styles[i][key] = last_known_values[key]

    return styles


def decode_old_text(unserialized: Dict) -> List[Dict]:
    """Decodes legacy text unpacked by NSKeyUnarchiver."""
    # string has already been processed
    # debug(f"here:{unserialized}")
    string = unserialized["NSString"]
    lengths = unserialized.get("NSAttributeInfo")
    styles = unserialized["NSAttributes"]

    formatted_data: List[Dict] = []

    # if text only contains one style
    if isinstance(styles, dict):
        lengths = [{"length": len(string), "attribute_id": 0}]
        styles = [styles]

    for length_info in lengths:
        length = length_info["length"]
        attribute_id = length_info["attribute_id"]

        # checking if NSAttributeInfo has successfully parsed
        if not styles or attribute_id < 0 or attribute_id >= len(styles):
            debug(
                f"Error: attribute_id {attribute_id} is out of range. styles length: {len(styles)}"
            )

        attribute = styles[attribute_id]

        # color
        # "NSColorSpace": 2 ?
        color_data = None
        ns_color = attribute.get("NSColor")
        if ns_color and isinstance(ns_color, dict):
            color_data = {
                "rgba": {
                    "red": ns_color.get("UIRed", 0),
                    "green": ns_color.get("UIGreen", 0),
                    "blue": ns_color.get("UIBlue", 0),
                    "alpha": ns_color.get("UIAlpha", 1),
                }
            }

        # paragraph
        paragraph_style = attribute.get("NSParagraphStyle")

        # font
        ns_font = attribute.get("NSFont")

        # TODO Include strokeStyle, lineHeight
        formatted_data.append(
            {
                # alignment might be wrong
                "alignment": paragraph_style.get("NSAlignment", 1) + 1,
                "length": length,
                "fillColor": color_data,
                "fontName": ns_font["NSName"],
                "fontSize": ns_font["NSSize"],
                "kerning": 0,
                "lineHeight": None,
                "strikethrough": False,
                "underline": False,
            }
        )

    return formatted_data



# coding=utf-8
#
# pylint: disable=invalid-name
#
"""
Depreicated simpletransform replacements with documentation
"""

import warnings

from inkex.deprecated import deprecate
from inkex.transforms import Transform, BoundingBox, cubic_extrema
from inkex.paths import Path

import inkex, cubicsuperpath


def _lists(mat):
    return [list(row) for row in mat]


@deprecate
def parseTransform(transf, mat=None):
    """Transform(str).matrix"""
    t = Transform(transf)
    if mat is not None:
        t = Transform(mat) @ t
    return _lists(t.matrix)


@deprecate
def formatTransform(mat):
    """str(Transform(mat))"""
    if len(mat) == 3:
        warnings.warn("3x3 matrices not suported")
        mat = mat[:2]
    return str(Transform(mat))


@deprecate
def invertTransform(mat):
    """-Transform(mat)"""
    return _lists((-Transform(mat)).matrix)


@deprecate
def composeTransform(mat1, mat2):
    """Transform(M1) * Transform(M2)"""
    return _lists((Transform(mat1) @ Transform(mat2)).matrix)


@deprecate
def composeParents(node, mat):
    """elem.composed_transform() or elem.transform * Transform(mat)"""
    return (node.transform @ Transform(mat)).matrix


@deprecate
def applyTransformToNode(mat, node):
    """elem.transform = Transform(mat) * elem.transform"""
    node.transform = Transform(mat) @ node.transform


@deprecate
def applyTransformToPoint(mat, pt):
    """Transform(mat).apply_to_point(pt)"""
    pt2 = Transform(mat).apply_to_point(pt)
    # Apply in place as original method was modifying arrays in place.
    # but don't do this in your code! This is not good code design.
    pt[0] = pt2[0]
    pt[1] = pt2[1]


@deprecate
def applyTransformToPath(mat, path):
    """Path(path).transform(mat)"""
    return Path(path).transform(Transform(mat)).to_arrays()


@deprecate
def fuseTransform(node):
    """node.apply_transform()"""
    return node.apply_transform()


@deprecate
def boxunion(b1, b2):
    """list(BoundingBox(b1) + BoundingBox(b2))"""
    bbox = BoundingBox(b1[:2], b1[2:]) + BoundingBox(b2[:2], b2[2:])
    return bbox.x.minimum, bbox.x.maximum, bbox.y.minimum, bbox.y.maximum


@deprecate
def roughBBox(path):
    """list(Path(path)).bounding_box())"""
    bbox = Path(path).bounding_box()
    return bbox.x.minimum, bbox.x.maximum, bbox.y.minimum, bbox.y.maximum


@deprecate
def refinedBBox(path):
    """list(Path(path)).bounding_box())"""
    bbox = Path(path).bounding_box()
    return bbox.x.minimum, bbox.x.maximum, bbox.y.minimum, bbox.y.maximum


@deprecate
def cubicExtrema(y0, y1, y2, y3):
    """from inkex.transforms import cubic_extrema"""
    return cubic_extrema(y0, y1, y2, y3)


@deprecate
def computeBBox(aList, mat=[[1, 0, 0], [0, 1, 0]]):
    """sum([node.bounding_box() for node in aList])"""
    return sum([node.bounding_box() for node in aList], None)


@deprecate
def computePointInNode(pt, node, mat=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]):
    """(-Transform(node.transform * mat)).apply_to_point(pt)"""
    return (-Transform(node.transform * mat)).apply_to_point(pt)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#                    Sergei Izmailov <sergei.a.izmailov@gmail.com>
#                    Windell Oskay <windell@oskay.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=attribute-defined-outside-init
#
"""
Provide a way to load lxml attributes with an svg API on top.
"""

import random
import math
from functools import cached_property

from lxml import etree

from ..interfaces.IElement import ISVGDocumentElement

from ..deprecated.meta import DeprecatedSvgMixin, deprecate
from ..units import discover_unit, parse_unit
from ._selected import ElementList
from ..transforms import BoundingBox
from ..styles import StyleSheets, ConditionalStyle

from ._base import BaseElement, ViewboxMixin
from ._meta import StyleElement, NamedView
from ._utils import registerNS, addNS, splitNS

from typing import Optional, List, Tuple

if False:  # pylint: disable=using-constant-test
    import typing  # pylint: disable=unused-import


class SvgDocumentElement(
    DeprecatedSvgMixin, ISVGDocumentElement, BaseElement, ViewboxMixin
):
    """Provide access to the document level svg functionality"""

    # pylint: disable=too-many-public-methods
    tag_name = "svg"

    selection: ElementList
    """The selection as passed by Inkscape (readonly)"""

    def _init(self):
        self.current_layer = None
        self.view_center = (0.0, 0.0)
        self.selection = ElementList(self)

    @cached_property
    def ids(self):
        result = {}
        for el in self.iter():
            try:
                id = super(etree.ElementBase, el).get("id", None)
                if id is not None:
                    result[id] = el

                el._root = self
            except TypeError:
                pass  # Comments
        return result

    @cached_property
    def stylesheet_cache(self):
        return {node: node.stylesheet() for node in self.xpath("//svg:style")}

    def tostring(self):
        """Convert document to string"""
        return etree.tostring(etree.ElementTree(self))

    def get_ids(self):
        """Returns a set of unique document ids"""
        return self.ids.keys()

    def get_unique_id(
        self,
        prefix: str,
        size: Optional[int] = None,
        blacklist: Optional[List[str]] = None,
    ):
        """Generate a new id from an existing old_id

        The id consists of a prefix and an appended random integer with size digits.

        If size is not given, it is determined automatically from the length of
        existing ids, i.e. those in the document plus those in the blacklist.

        Args:
            prefix (str): the prefix of the new ID.
            size (Optional[int], optional): number of digits of the second part of the
                id. If None, the length is chosen based on the amount of existing
                objects. Defaults to None.

                .. versionchanged:: 1.1
                    The default of this parameter has been changed from 4 to None.
            blacklist (Optional[Iterable[str]], optional): An additional iterable of ids
                that are not allowed to be used. This is useful when bulk inserting
                objects.
                Defaults to None.

                .. versionadded:: 1.2

        Returns:
            _type_: _description_
        """
        ids = self.get_ids()
        if size is None:
            size = max(math.ceil(math.log10(len(ids) or 1000)) + 1, 4)
        new_id = None
        _from = 10**size - 1
        _to = 10**size
        while (
            new_id is None
            or new_id in ids
            or (blacklist is not None and new_id in blacklist)
        ):
            # Do not use randint because py2/3 incompatibility
            new_id = prefix + str(int(random.random() * _from - _to) + _to)
        return new_id

    def get_page_bbox(self, page=None) -> BoundingBox:
        """Gets the page dimensions as a bbox. For single-page documents, the viewbox
        dimensions are returned.

        Args:
            page (int, optional): Page number. Defaults to the first page.

                .. versionadded:: 1.3

        Raises:
            IndexError: if the page number provided does not exist in the document.

        Returns:
            BoundingBox: the bounding box of the page
        """
        if page is None:
            page = 0
        pages = self.namedview.get_pages()
        if 0 <= page < len(pages):
            return pages[page].bounding_box
        raise IndexError("Invalid page number")

    def get_current_layer(self):
        """Returns the currently selected layer"""
        layer = self.getElementById(self.namedview.current_layer, "svg:g")
        if layer is None:
            return self
        return layer

    def add_namespace(self, prefix, url):
        """Adds an xml namespace to the xml parser with the desired prefix.

        If the prefix or url are already in use with different values, this
        function will raise an error. Remove any attributes or elements using
        this namespace before calling this function in order to rename it.

        .. versionadded:: 1.3
        """
        if self.nsmap.get(prefix, None) == url:
            registerNS(prefix, url)
            return

        # Attempt to clean any existing namespaces
        if prefix in self.nsmap or url in self.nsmap.values():
            nskeep = [k for k, v in self.nsmap.items() if k != prefix and v != url]
            etree.cleanup_namespaces(self, keep_ns_prefixes=nskeep)
            if prefix in self.nsmap:
                raise KeyError("ns prefix already used with a different url")
            if url in self.nsmap.values():
                raise ValueError("ns url already used with a different prefix")

        # These are globals, but both will overwrite previous uses.
        registerNS(prefix, url)
        etree.register_namespace(prefix, url)

        # Set and unset an attribute to add the namespace to this root element.
        self.set(f"{prefix}:temp", "1")
        self.set(f"{prefix}:temp", None)

    def getElement(self, xpath):  # pylint: disable=invalid-name
        """Gets a single element from the given xpath or returns None"""
        return self.findone(xpath)

    def getElementById(self, eid: str, elm="*", literal=False):  # pylint: disable=invalid-name
        """Get an element in this svg document by it's ID attribute.

        Args:
            eid (str): element id
            elm (str, optional): element type, including namespace, e.g. ``svg:path``.
                Defaults to "*".
            literal (bool, optional): If ``False``, ``#url()`` is stripped from ``eid``.
                Defaults to False.

                .. versionadded:: 1.1

        Returns:
            Union[BaseElement, None]: found element
        """
        if eid is not None and not literal:
            eid = eid.strip()[4:-1] if eid.startswith("url(") else eid
            eid = eid.lstrip("#")

        result = self.ids.get(eid, None)
        if result is not None:
            if elm != "*":
                elm_with_ns = addNS(*splitNS(elm)[::-1])
                if not super(etree.ElementBase, result).tag == elm_with_ns:
                    return None
            return result
        return None

    def getElementByName(self, name, elm="*"):  # pylint: disable=invalid-name
        """Get an element by it's inkscape:label (aka name)"""
        return self.getElement(f'//{elm}[@inkscape:label="{name}"]')

    def getElementsByClass(self, class_name):  # pylint: disable=invalid-name
        """Get elements by it's class name"""
        return ConditionalStyle(f".{class_name}").all_matches(self)

    def getElementsByHref(self, eid: str, attribute="href"):  # pylint: disable=invalid-name
        """Get elements that reference the element with id eid.

        Args:
            eid (str): _description_
            attribute (str, optional): Attribute to look for.
                Valid choices: "href", "xlink:href", "mask", "clip-path".
                Defaults to "href".

                .. versionadded:: 1.2

            attribute set to "href" or "xlink:href" handles both cases.
                .. versionchanged:: 1.3

        Returns:
            Any: list of elements
        """
        if attribute == "href" or attribute == "xlink:href":
            return self.xpath(f'//*[@href|@xlink:href="#{eid}"]')
        elif attribute == "mask":
            return self.xpath(f'//*[@mask="url(#{eid})"]')
        elif attribute == "clip-path":
            return self.xpath(f'//*[@clip-path="url(#{eid})"]')

    def getElementsByStyleUrl(self, eid, style=None):  # pylint: disable=invalid-name
        """Get elements by a style attribute url"""
        url = f"url(#{eid})"
        if style is not None:
            url = style + ":" + url
        return self.xpath(f'//*[contains(@style,"{url}")]')

    @property
    def name(self):
        """Returns the Document Name"""
        return self.get("sodipodi:docname", "")

    @property
    def namedview(self) -> NamedView:
        """Return the sp namedview meta information element"""
        return self.get_or_create("//sodipodi:namedview", prepend=True)

    @property
    def metadata(self):
        """Return the svg metadata meta element container"""
        return self.get_or_create("//svg:metadata", prepend=True)

    @property
    def defs(self):
        """Return the svg defs meta element container"""
        return self.get_or_create("//svg:defs", prepend=True)

    def get_viewbox(self) -> List[float]:
        """Parse and return the document's viewBox attribute"""
        return self.parse_viewbox(self.get("viewBox", "0")) or [0, 0, 0, 0]

    @property
    def viewbox_width(self) -> float:  # getDocumentWidth(self):
        """Returns the width of the `user coordinate system
        <https://www.w3.org/TR/SVG2/coords.html#Introduction>`_ in user units, i.e.
        the width of the viewbox, as defined in the SVG file. If no viewbox is defined,
        the value of the width attribute is returned. If the height is not defined,
        returns 0.

        .. versionadded:: 1.2
        """
        return self.get_viewbox()[2] or self.viewport_width

    @property
    def viewport_width(self) -> float:
        """Returns the width of the `viewport coordinate system
        <https://www.w3.org/TR/SVG2/coords.html#Introduction>`_ in user units, i.e. the
        width attribute of the svg element converted to px

        .. versionadded:: 1.2
        """
        return self.to_dimensionless(self.get("width")) or self.get_viewbox()[2]

    @property
    def viewbox_height(self) -> float:  # getDocumentHeight(self):
        """Returns the height of the `user coordinate system
        <https://www.w3.org/TR/SVG2/coords.html#Introduction>`_ in user units, i.e. the
        height of the viewbox, as defined in the SVG file. If no viewbox is defined, the
        value of the height attribute is returned. If the height is not defined,
        returns 0.

        .. versionadded:: 1.2
        """
        return self.get_viewbox()[3] or self.viewport_height

    @property
    def viewport_height(self) -> float:
        """Returns the width of the `viewport coordinate system
        <https://www.w3.org/TR/SVG2/coords.html#Introduction>`_ in user units, i.e. the
        height attribute of the svg element converted to px

        .. versionadded:: 1.2
        """
        return self.to_dimensionless(self.get("height")) or self.get_viewbox()[3]

    @property
    def scale(self):
        """Returns the ratio between the viewBox width and the page width.

        .. versionchanged:: 1.2
            Previously, the scale as shown by the document properties was computed,
            but the computation of this in core Inkscape changed in Inkscape 1.2, so
            this was moved to :attr:`inkscape_scale`.
        """
        return self._base_scale()

    @property
    def inkscape_scale(self):
        """Returns the ratio between the viewBox width (in width/height units) and the
        page width, which is displayed as "scale" in the Inkscape document
        properties.

        .. versionadded:: 1.2
        """

        viewbox_unit = (
            parse_unit(self.get("width")) or parse_unit(self.get("height")) or (0, "px")
        )[1]
        return self._base_scale(viewbox_unit)

    def _base_scale(self, unit="px"):
        """Returns what Inkscape shows as "user units per `unit`"

        .. versionadded:: 1.2
        """
        try:
            scale_x = (
                self.to_dimensional(self.viewport_width, unit) / self.viewbox_width
            )
            scale_y = (
                self.to_dimensional(self.viewport_height, unit) / self.viewbox_height
            )
            value = max([scale_x, scale_y])
            return 1.0 if value == 0 else value
        except (ValueError, ZeroDivisionError):
            return 1.0

    @property
    def equivalent_transform_scale(self) -> float:
        """Return the scale of the equivalent transform of the svg tag, as defined by
        https://www.w3.org/TR/SVG2/coords.html#ComputingAViewportsTransform
        (highly simplified)

        .. versionadded:: 1.2
        """
        return self.scale

    @property
    def unit(self):
        """Returns the unit used for in the SVG document.

        In the case the SVG document lacks an attribute that explicitly
        defines what units are used for SVG coordinates, it tries to calculate
        the unit from the SVG width and viewBox attributes.
        Defaults to 'px' units.
        """
        if not hasattr(self, "_unit"):
            self._unit = "px"  # Default is px
            viewbox = self.get_viewbox()
            if viewbox and set(viewbox) != {0}:
                self._unit = discover_unit(self.get("width"), viewbox[2], default="px")
        return self._unit

    @property
    def document_unit(self):
        """Returns the display unit (Inkscape-specific attribute) of the document

        .. versionadded:: 1.2
        """
        return self.namedview.get("inkscape:document-units", "px")

    @property
    def stylesheets(self):
        """Get all the stylesheets, bound together to one, (for reading)"""
        sheets = StyleSheets()
        for value in self.stylesheet_cache.values():
            sheets.append(value)
        return sheets

    @property
    def stylesheet(self):
        """Return the first stylesheet or create one if needed (for writing)"""
        for sheet in self.stylesheets:
            return sheet

        style_node = StyleElement()
        self.defs.append(style_node)
        return style_node.stylesheet()

    def add_to_tree_callback(self, element):
        """Callback called automatically when adding an element to the tree.
        Updates the list of stylesheets and the ID tracker with the subtree of element.

        .. versionadded:: 1.4

        Args:
            element (BaseElement): element added to the tree.
        """

        for el in element.iter():
            self._add_individual_to_tree(el)

    def _add_individual_to_tree(self, element: BaseElement):
        if isinstance(element, etree._Comment):
            return
        element._root = self
        if element.TAG == "style":
            self.stylesheet_cache[element] = element.stylesheet()
        new_id = element.get("id", None)
        if new_id is not None:
            if new_id in self.ids:
                while new_id in self.ids:
                    new_id += "-1"
                super(etree.ElementBase, element).set("id", new_id)  # type: ignore
            self.ids[new_id] = element

    def remove_from_tree_callback(self, element):
        """Callback called automatically when removing an element from the tree.
        Remove elements in the subtree of element from the the list of stylesheets
        and the ID tracker.

        .. versionadded:: 1.4

        Args:
            element (BaseElement): element added to the tree.
        """
        for el in element.iter():
            self._remove_individual_from_tree(el)

    def _remove_individual_from_tree(self, element):
        if isinstance(element, etree._Comment):
            return
        element._root = None
        if element.TAG == "style" and element in self.stylesheet_cache:
            self.stylesheet_cache.remove(element)
        old_id = element.get("id", None)
        if old_id is not None and old_id in self.ids:
            self.ids.pop(old_id)


def width(self):
    """Use :func:`viewport_width` instead"""
    return self.viewport_width


def height(self):
    """Use :func:`viewport_height` instead"""
    return self.viewport_height


SvgDocumentElement.width = property(deprecate(width, "1.2"))
SvgDocumentElement.height = property(deprecate(height, "1.2"))



# coding=utf-8
#
# Copyright (C) 2010 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Python barcode renderer for UPCE barcodes. Designed for use with Inkscape.
"""

from .BaseEan import EanBarcode
from inkex.localization import inkex_gettext as _

# This is almost exactly the same as the standard FAMILIES
# But flipped around and with the first 111000 instead of 000000.
FAMS = [
    "111000",
    "110100",
    "110010",
    "110001",
    "101100",
    "100110",
    "100011",
    "101010",
    "101001",
    "100101",
]


class Upce(EanBarcode):
    """Generate EAN6/UPC-E barcode generator"""

    name = "upce"
    font_size = 10
    lengths = [6, 11]
    checks = [7, 12]
    center_bar = "020"

    def _encode(self, num, guide=False):
        """Generate a UPC-E Barcode"""
        self.text = EanBarcode.space(["0"], 2, num[:6], 2, num[-1])
        code = EanBarcode.encode_interleaved(num[-1], num[:6], FAMS)
        return self.enclose(code)

    def append_checksum(self, number):
        """Generate a UPCE Checksum"""
        if len(number) == 6:
            number = Upce.convert_e2a(number)
        result = self.get_checksum(number)
        return Upce.convert_a2e(number) + result

    @staticmethod
    def convert_a2e(number):
        """Converting UPC-A to UPC-E, may cause errors."""
        # All UPC-E Numbers use number system 0
        if number[0] != "0" or len(number) != 11:
            # If not then the code is invalid
            raise ValueError(_("Invalid UPC Number"))

        # Most of the conversions deal
        # with the specific code parts
        maker = number[1:6]
        product = number[6:11]

        # There are 4 cases to convert:
        if maker[2:] == "000" or maker[2:] == "100" or maker[2:] == "200":
            # Maximum number product code digits can be encoded
            if product[:2] == "00":
                return maker[:2] + product[2:] + maker[2]
        elif maker[3:5] == "00":
            # Now only 2 product code digits can be used
            if product[:3] == "000":
                return maker[:3] + product[3:] + "3"
        elif maker[4] == "0":
            # With even more maker code we have less room for product code
            if product[:4] == "0000":
                return maker[0:4] + product[4] + "4"
        elif product[:4] == "0000" and int(product[4]) > 4:
            # The last recorse is to try and squeeze it in the last 5 numbers
            # so long as the product is 00005-00009 so as not to conflict with
            # the 0-4 used above.
            return maker + product[4]
        # Invalid UPC-A Numbe
        raise ValueError(_("Invalid UPC Number"))

    @staticmethod
    def convert_e2a(number):
        """Convert UPC-E to UPC-A by padding with zeros"""
        # It's more likly to convert this without fault
        # But we still must be mindful of the 4 conversions
        if len(number) != 6:
            return None

        if number[5] in ["0", "1", "2"]:
            return "0" + number[:2] + number[5] + "0000" + number[2:5]
        if number[5] == "3":
            return "0" + number[:3] + "00000" + number[3:5]
        if number[5] == "4":
            return "0" + number[:4] + "00000" + number[4]
        return "0" + number[:5] + "0000" + number[5]



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
# SPDX-FileCopyrightText: 2024 Software Freedom Conservancy <info@sfconservancy.org>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from enum import Enum
from typing import Callable, Dict, List, Optional

import inkex

from ..parser.types import AFObject
from .util import interpolate_float


class AFColorSpace(Enum):
    UNKNOWN = 0
    RGB = 1
    HSL = 2
    CMYK = 3
    LAB = 4
    GRAY = 5

    @classmethod
    def from_string(cls, space: str) -> AFColorSpace:
        if space.upper() in ["RGB", "RGBA"]:
            return AFColorSpace.RGB
        if space.upper() in ["HSL", "HSLA"]:
            return AFColorSpace.HSL
        if space.upper() == "CMYK":
            return AFColorSpace.CMYK
        if space.upper() in ["LAB", "LABA"]:
            return AFColorSpace.LAB
        if space.upper() == "GRAY":
            return AFColorSpace.GRAY
        return AFColorSpace.UNKNOWN


class AFGradientType(Enum):
    LINEAR = 0
    ELLIPTICAL = 1
    RADIAL = 2
    CONICAL = 3


def gray_to_rgb(color: AFColor) -> AFColor:
    assert color.space == AFColorSpace.GRAY
    return AFColor([color.color[0]] * 3, alpha=color.alpha, _space=AFColorSpace.RGB)


def hsl_to_rgb(color: AFColor) -> AFColor:
    assert color.space == AFColorSpace.HSL
    return AFColor(
        inkex.colors.hsl_to_rgb(*color.color),
        alpha=color.alpha,
        _space=AFColorSpace.RGB,
    )


def cmyk_to_rgb(color: AFColor) -> AFColor:
    assert color.space == AFColorSpace.CMYK
    c, m, y, k = color.color
    if k == 1:
        return AFColor([0, 0, 0], alpha=color.alpha, _space=AFColorSpace.RGB)
    return AFColor(
        [
            0.0 if c >= 1 else 1.0 * (1 - c) * (1 - k),
            0.0 if m >= 1 else 1.0 * (1 - m) * (1 - k),
            0.0 if y >= 1 else 1.0 * (1 - y) * (1 - k),
        ],
        alpha=color.alpha,
        _space=AFColorSpace.RGB,
    )


def lab_to_rgb(color: AFColor) -> AFColor:
    raise NotImplementedError("LAB to RGB not implemented")


map_to_rgb: Dict[AFColorSpace, Callable[[AFColor], AFColor]] = {
    AFColorSpace.RGB: lambda color: color,
    AFColorSpace.HSL: hsl_to_rgb,
    AFColorSpace.CMYK: cmyk_to_rgb,
    AFColorSpace.LAB: lab_to_rgb,
    AFColorSpace.GRAY: gray_to_rgb,
}

map_from_rgb: Dict[AFColorSpace, Callable[[AFColor], AFColor]] = {
    AFColorSpace.RGB: lambda color: color,
    # AFColorSpace.HSL: lambda color: color,
    # AFColorSpace.CMYK: lambda color: color,
    # AFColorSpace.LAB: lambda color: color,
    # AFColorSpace.GRAY: lambda color: color,
}


@dataclass
class AFColor:
    color: list[int | float]
    alpha: float
    _space: InitVar[AFColorSpace | str]
    space: AFColorSpace = field(init=False)

    def __post_init__(self, _space: AFColorSpace | str):
        self.space = (
            _space
            if isinstance(_space, AFColorSpace)
            else AFColorSpace.from_string(_space)
        )
        assert self.space != AFColorSpace.UNKNOWN, f"Unknown color space: {_space}"

        assert 0 <= self.alpha <= 1, f"Bad alpha: {self.alpha}"
        assert len(self.color) == 1 or self.space != AFColorSpace.GRAY, (
            f"Bad {self.space.name} color list: {self.color}"
        )
        assert len(self.color) == 3 or self.space not in [
            AFColorSpace.RGB,
            AFColorSpace.HSL,
            AFColorSpace.LAB,
        ], f"Bad {self.space.name} color list: {self.color}"
        assert len(self.color) == 4 or self.space != AFColorSpace.CMYK, (
            f"Bad {self.space.name} color list: {self.color}"
        )

    @classmethod
    def from_af(cls, color: AFObject) -> AFColor:
        color_values = color["_col"]
        color_space = color.get_type()

        if color_space == "RGBA":
            return AFColor(
                color_values[:3], alpha=color_values[3], _space=AFColorSpace.RGB
            )
        if color_space == "HSLA":
            return AFColor(
                color_values[:3], alpha=color_values[3], _space=AFColorSpace.HSL
            )
        if color_space == "CMYK":
            return AFColor(
                color_values[:4], alpha=color_values[4], _space=AFColorSpace.CMYK
            )
        if color_space == "LABA":
            return AFColor(
                color_values[:3],
                alpha=color_values[3] / 0xFFFF,
                _space=AFColorSpace.LAB,
            )
        if color_space == "GRAY":
            return AFColor(
                [color_values[0]], alpha=color_values[1], _space=AFColorSpace.GRAY
            )

        raise NotImplementedError(f"Unknown color space: {color_space}")

    def to_rgb(self) -> AFColor:
        fn = map_to_rgb.get(self.space)
        if fn is None:
            raise NotImplementedError(f"Can't convert {self.space.name} to RGB")
        return fn(self)

    def to_space(self, space: AFColorSpace) -> AFColor:
        if self.space != AFColorSpace.RGB:
            return AFColor(
                self.to_rgb().to_space(space).color, alpha=self.alpha, _space=space
            )
        fn = map_from_rgb.get(space)
        if fn is None:
            raise NotImplementedError(
                f"Can't convert {self.space.name} to {space.name}"
            )
        return fn(self)

    def interpolate(self, other: AFColor, t: float, space=AFColorSpace.RGB) -> AFColor:
        inter = inkex.Color(
            self.to_space(space).color, space=space.name.lower()
        ).interpolate(
            inkex.Color(other.to_space(space).color, space=space.name.lower()), t
        )
        return AFColor(
            [c / 255 for c in inter],
            alpha=interpolate_float(self.alpha, other.alpha, t),
            _space=inter.space,
        )


@dataclass
class AFGradientStop:
    pos: float
    mid: float
    color: AFColor


@dataclass
class AFGradient:
    grad_type: AFGradientType
    tfm: inkex.Transform
    stops: List[AFGradientStop]
    fdsc: bool = True

    @classmethod
    def from_af(
        cls, fdef: AFObject, fdex: List[float], fdsc: Optional[bool]
    ) -> AFGradient:
        assert fdef.get_type() == "FilG", "Not a gradient"
        grad_type = AFGradientType(fdef["Type"].id)

        fdsc = True if fdsc is None else fdsc
        assert len(fdex) == 6, "Invalid gradient transform"
        assert isinstance(fdsc, bool)
        if not fdsc:
            raise NotImplementedError(f"FDSc {fdsc} not implemented")  # TODO

        grad = fdef["Grad"]
        stops = [
            AFGradientStop(s[0], s[1], AFColor.from_af(c))
            for s, c in zip(grad["Posn"], grad["Cols"])
        ]

        tfm = inkex.Transform((fdex[:3], fdex[3:]))  # type: ignore
        return cls(grad_type, tfm, stops, fdsc)


def parse_pant(fill: AFObject) -> AFColor:
    # TODO: fields: ispt, name, base, srgb
    return AFColor.from_af(fill["srgb"])


def parse_solid(fill: AFObject) -> AFColor:
    # TODO: type: RegC
    if fill["Colr"].get_type() == "Pant":
        return parse_pant(fill["Colr"])
    return AFColor.from_af(fill["Colr"])


def parse_no_fill(_: Optional[AFObject] = None) -> None:
    return None


def parse_gradient(
    fill: AFObject, tfm: List[float], fdsc: Optional[bool] = None
) -> AFGradient:
    return AFGradient.from_af(fill, tfm, fdsc)


fdsc_map: Dict[str, Callable[[AFObject], Optional[AFColor | AFGradient]]] = {
    "FilN": lambda _: parse_no_fill(),
    "FilS": lambda fdsc: parse_solid(fdsc["FDeF"]),
    "FilG": lambda fdsc: parse_gradient(fdsc["FDeF"], fdsc["FDeX"], fdsc.get("FDSc")),
    # "FilB": # TODO,
}


def parse_fdsc(fdsc: AFObject) -> Optional[AFColor | AFGradient]:
    assert fdsc.get_type() == "FDsc", f"Not a fill desc: {fdsc}"
    fill_type = fdsc["FDeF"].get_type()
    assert fill_type in [
        "FilN",
        "FilS",
        "FilG",
        "FilB",
    ], f"Invalid fill type: {fill_type}"

    fn = fdsc_map.get(fill_type)
    if fn is None:
        raise NotImplementedError(f"Fill type '{fill_type}' not implemented")

    return fn(fdsc)


def parse_bfil(bfil: AFObject) -> Optional[AFColor | AFGradient]:
    if bfil.get_type() == "FDsc":
        return parse_fdsc(bfil)
    if bfil.get_type() == "FilS":
        return parse_solid(bfil)
    raise NotImplementedError(f"Fill type '{bfil.get_type()}' not implemented")



#
# Copyright 2021 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Access Reactome Content Services.
"""

import re

from import_sources import RemoteSource

TAG_REX = re.compile(r"<[^<]+?>")


class Reactome(RemoteSource):
    name = "Reactome (Bio)"
    icon = "sources/reactome.svg"
    search_url = "https://reactome.org/ContentService/search/query"
    file_url = "https://reactome.org/icon/{stId}.svg"
    icon_url = "https://reactome.org/icon/{stId}.png"
    all_licence = "cc-by-sa-4.0"

    def search(self, query):
        params = {
            "query": query,
            "types": "Icon",
            "cluster": "true",
            "Start row": 0,
            "rows": 100,
        }
        response = {}
        try:
            response = self.session.get(self.search_url, params=params).json()
        except Exception:
            pass

        if "messages" in response and "No entries" in response["messages"][0]:
            return
        for cats in response.get("results", []):
            for entry in cats["entries"]:
                yield {
                    "id": entry["dbId"],
                    "name": TAG_REX.sub("", entry["name"]),
                    "author": "Reactome/" + entry.get("iconDesignerName", "Unknown"),
                    "summary": TAG_REX.sub("", entry.get("summation", "")),
                    "created": None,  # No data
                    "popularity": 0,  # No data
                    "thumbnail": self.icon_url.format(**entry),
                    "file": self.file_url.format(**entry),
                    "license": self.all_licence,
                }



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2010 Craig Marshall, craig9 [at] gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
"""
This script finds all fonts in the current drawing that match the
specified find font, and replaces them with the specified replacement
font.

It can also replace all fonts indiscriminately, and list all fonts
currently being used.
"""

import inkex
from inkex.localization import inkex_gettext as _

text_tags = [
    "{http://www.w3.org/2000/svg}tspan",
    "{http://www.w3.org/2000/svg}text",
    "{http://www.w3.org/2000/svg}flowRoot",
    "{http://www.w3.org/2000/svg}flowPara",
    "{http://www.w3.org/2000/svg}flowSpan",
]
font_attributes = ["font-family", "-inkscape-font-specification"]


def set_font(node, new_font, style=None):
    """
    Sets the font attribute in the style attribute of node, using the
    font name stored in new_font. If the style dict is open already,
    it can be passed in, otherwise it will be optned anyway.

    Returns a dirty boolean flag
    """
    dirty = False
    if not style:
        style = node.style
    if style:
        for att in font_attributes:
            if att in style:
                style[att] = new_font
                node.style = style
                dirty = True
    return dirty


def find_replace_font(node, find, replace):
    """
    Searches the relevant font attributes/styles of node for find, and
    replaces them with replace.

    Returns a dirty boolean flag
    """
    dirty = False
    style = node.style
    if style:
        for att in font_attributes:
            if style(att).strip().lower() == find:
                set_font(node, replace, style)
                dirty = True
    return dirty


def is_styled_text(node):
    """
    Returns true if the tag in question is a "styled" element that
    can hold text.
    """
    return node.tag in text_tags and "style" in node.attrib


def is_text(node):
    """
    Returns true if the tag in question is an element that
    can hold text.
    """
    return node.tag in text_tags


def get_fonts(node):
    """
    Given a node, returns a list containing all the fonts that
    the node is using.
    """
    fonts = []
    s = node.style
    if not s:
        return fonts
    for a in font_attributes:
        if a in s:
            fonts.append(s[a])
    return fonts


def report_replacements(num):
    """
    Sends a message to the end user showing success of failure
    of the font replacement
    """
    if num == 0:
        inkex.errormsg(
            _(
                "Couldn't find anything using that font, please ensure the spelling "
                "and spacing is correct."
            )
        )


def report_findings(findings):
    """
    Tells the user which fonts were found, if any
    """
    if len(findings) == 0:
        inkex.errormsg(_("Didn't find any fonts in this document/selection."))
    else:
        if len(findings) == 1:
            inkex.errormsg(_("Found the following font only: %s") % findings[0])
        else:
            inkex.errormsg(_("Found the following fonts:\n%s") % "\n".join(findings))


class ReplaceFont(inkex.EffectExtension):
    """
    Replaces all instances of one font with another
    """

    def add_arguments(self, pars):
        pars.add_argument("--fr_find")
        pars.add_argument("--fr_replace")
        pars.add_argument("--r_replace")
        pars.add_argument("--action")
        pars.add_argument("--scope")

    def find_child_text_items(self, node):
        """
        Recursive method for appending all text-type elements
        to self.selected_items
        """
        if is_text(node):
            yield node

        for child in node:
            for textchild in self.find_child_text_items(child):
                yield textchild

    def relevant_items(self, scope):
        """
        Depending on the scope, returns all text elements, or all
        selected text elements including nested children
        """
        items = []
        to_return = []

        selected = self.svg
        if scope == "selection_only":
            selected = self.svg.selection.values()

        for item in selected:
            items.extend(self.find_child_text_items(item))

        if not items:
            return inkex.errormsg(_("There was nothing selected"))

        return items

    def find_replace(self, nodes, find, replace):
        """
        Walks through nodes, replacing fonts as it goes according
        to find and replace
        """
        replacements = 0
        for node in nodes:
            if find_replace_font(node, find, replace):
                replacements += 1
        report_replacements(replacements)

    def replace_all(self, nodes, replace):
        """
        Walks through nodes, setting fonts indiscriminately.
        """
        replacements = 0
        for node in nodes:
            if set_font(node, replace):
                replacements += 1
        report_replacements(replacements)

    def list_all(self, nodes):
        """
        Walks through nodes, building a list of all fonts found, then
        reports to the user with that list
        """
        fonts_found = []
        for node in nodes:
            for f in get_fonts(node):
                if not f in fonts_found:
                    fonts_found.append(f)
        report_findings(sorted(fonts_found))

    def effect(self):
        if not self.options.action:
            return inkex.errormsg(_("Nothing to do, no action specified."))
        action = self.options.action.strip(
            '"'
        )  # TODO Is this a bug? (Extra " characters)
        scope = self.options.scope

        relevant_items = self.relevant_items(scope)

        if action == "find_replace":
            find = self.options.fr_find
            if find is None or find == "":
                return inkex.errormsg(
                    _("Please enter a search string in the find box.")
                )
            find = find.strip().lower()
            replace = self.options.fr_replace
            if replace is None or replace == "":
                return inkex.errormsg(
                    _("Please enter a replacement font in the replace with box.")
                )
            self.find_replace(relevant_items, find, replace)
        elif action == "replace_all":
            replace = self.options.r_replace
            if replace is None or replace == "":
                return inkex.errormsg(
                    _("Please enter a replacement font in the replace all box.")
                )
            self.replace_all(relevant_items, replace)
        elif action == "list_only":
            self.list_all(relevant_items)


if __name__ == "__main__":
    ReplaceFont().run()



# coding=utf-8
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=invalid-name
"""Deprecated cubic super path API"""

from inkex.deprecated import deprecate
from inkex import paths


@deprecate
def ArcToPath(p1, params):
    return paths.arc_to_path(p1, params)


@deprecate
def CubicSuperPath(simplepath):
    return paths.Path(simplepath).to_superpath()


@deprecate
def unCubicSuperPath(csp):
    return paths.CubicSuperPath(csp).to_path().to_arrays()


@deprecate
def parsePath(d):
    return paths.CubicSuperPath(paths.Path(d))


@deprecate
def formatPath(p):
    return str(paths.Path(unCubicSuperPath(p)))


matprod = deprecate(paths.matprod)
rotmat = deprecate(paths.rotmat)
applymat = deprecate(paths.applymat)
norm = deprecate(paths.norm)



#!/usr/bin/env python3

# Copyright (C) 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Constants for the parser module
"""

file_magic = 0x414BFF00
content_magic = 0x534BFF00

expected_file_classes = [
    "Prsn",
    "Pref",
    "Adjm",
    "AstP",
    "CroP",
    "DevP",
    "Dcts",
    "DocP",
    "ExpP",
    "Fils",
    "GrdP",
    "Macs",
    "Objs",
    "OSty",
    "RBru",
    "Shps",
    "TonP",
    "TlSt",
    "VBru",
]

expected_documents = [
    "doc.dat",
    "newdoccontroller",
    "previews.dat",
    "preferences",
    "newdocstate",
    "session.dat",
    "structure",
    "content.dat",
    "templates.dat",
]



#
# Copyright 2011-2022 Martin Owens <doctormo@geek-2.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
# pylint: disable=wrong-import-position
"""
This is a wrapper layer to make interacting with Gtk a little less painful.
The main issues with Gtk is that it expects an awful lot of the developer,
code which is repeated over and over and patterns which every single developer
will use are not given easy to use convenience functions.

This makes Gtk programming WET, unattractive and error prone. This module steps
inbetween and adds in all those missing bits. It's not meant to replace Gtk and
certainly it's possible to use Gtk and threading directly.

.. versionadded:: 1.2
"""

import os
import sys
import logging
import threading

from ..utils import DependencyError

try:
    import gi

    gi.require_version("Gtk", "3.0")

    # Importing while covering stderr because pygobject has broken
    # warnings support and will force import warnings on our users.
    tmp, sys.stderr = sys.stderr, None  # type: ignore
    from gi.repository import Gtk, GLib

    sys.stderr = tmp  # type: ignore
except ImportError:  # pragma: no cover
    raise DependencyError(
        "You are missing the required libraries for Gtk."
        " Please report this problem to the Inkscape developers."
    )

from .app import GtkApp
from .window import Window, ChildWindow, FakeWidget
from .listview import TreeView, IconView, ViewColumn, ViewSort, Separator
from .pixmap import PixmapManager



# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
# Copyright (C) 2023 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Quadratic and TepidQuadratic path commands"""

from __future__ import annotations

from typing import overload, Tuple, Callable, Union
from math import sqrt

import numpy as np

from ..transforms import quadratic_extrema, Transform, ComplexLike

from .interfaces import (
    AbsolutePathCommand,
    RelativePathCommand,
    BezierComputationMixin,
    BezierArcComputationMixin,
    LengthSettings,
)


class QuadraticMixin(BezierComputationMixin, BezierArcComputationMixin):
    # pylint: disable=unused-argument
    ccontrol_points: Callable[[complex, complex, complex], Tuple[complex, ...]]

    def _cderivative(
        self, first: complex, prev: complex, prev_control: complex, t: float, n: int = 1
    ) -> complex:
        points = self.ccontrol_points(first, prev, prev_control)
        if n == 1:
            return 2 * ((points[0] - prev) * (1 - t) + (points[1] - points[0]) * t)
        if n == 2:
            return 2 * (prev - 2 * points[0] + points[1])
        if n > 2:
            return 0j
        raise ValueError("n should be a positive integer.")

    def _cunit_tangent(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        return self.bezier_unit_tangent(prev, prev_control, t)

    def _cpoint(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        control, end = self.ccontrol_points(first, prev, prev_control)
        return (1 - t) ** 2 * prev + 2 * t * (1 - t) * control + t**2 * end

    # TODO maybe better treatment of degenerate beziers from
    # https://github.com/linebender/kurbo/blob/c229a914d303c5989c9e6b1d766def2df27a8185/src/quadbez.rs#L239
    # Ported from https://github.com/mathandy/svgpathtools/blob/19df25b99b405ec4fc7616b58384eca7879b6fd4/svgpathtools/path.py#L919
    # (MIT licensed)
    def _length(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        t0: float = 0,
        t1: float = 1,
        settings=LengthSettings(),
    ) -> float:
        control, end = self.ccontrol_points(first, prev, prev_control)
        a = prev - 2 * control + end
        b = 2 * (control - prev)

        if abs(a) < 1e-12:
            s = abs(b) * (t1 - t0)
        else:
            c2 = 4 * (a.real**2 + a.imag**2)
            c1 = 4 * (a.real * b.real + a.imag * b.imag)
            c0 = b.real**2 + b.imag**2

            beta = c1 / (2 * c2)
            gamma = c0 / c2 - beta**2

            dq1_mag = sqrt(c2 * t1**2 + c1 * t1 + c0)
            dq0_mag = sqrt(c2 * t0**2 + c1 * t0 + c0)
            # this implicitly handles division by zero
            try:
                logarand = (sqrt(c2) * (t1 + beta) + dq1_mag) / (
                    sqrt(c2) * (t0 + beta) + dq0_mag
                )

                s = (
                    (t1 + beta) * dq1_mag
                    - (t0 + beta) * dq0_mag
                    + gamma * sqrt(c2) * np.log(logarand)
                ) / 2
            except ZeroDivisionError:
                s = np.nan
            if np.isnan(s):
                tstar = abs(b) / (2 * abs(a))
                if t1 < tstar:
                    return abs(a) * (t0**2 - t1**2) - abs(b) * (t0 - t1)
                elif tstar < t0:
                    return abs(a) * (t1**2 - t0**2) - abs(b) * (t1 - t0)
                else:
                    return (
                        abs(a) * (t1**2 + t0**2)
                        - abs(b) * (t1 + t0)
                        + abs(b) ** 2 / (2 * abs(a))
                    )
        return s

    def _abssplit(
        self, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Quadratic, Quadratic]:
        """Split this Quadratic and return two Quadratics using DeCasteljau's algorithm"""
        p1, p2 = self.ccontrol_points(0j, prev, prev_control)
        p1_1 = (1 - t) * prev + t * p1
        p1_2 = (1 - t) * p1 + t * p2
        p2_1 = (1 - t) * p1_1 + t * p1_2
        return Quadratic(p1_1, p2_1), Quadratic(p1_2, p2)

    def _relsplit(self, prev: complex, prev_control: complex, t: float):
        """Split this curve and return two curves"""
        c1abs, c2abs = self._abssplit(prev, prev_control, t)
        return c1abs.to_relative(prev), c2abs.to_relative(c1abs.arg2)


class Quadratic(QuadraticMixin, AbsolutePathCommand):
    """Absolute Quadratic Curved Line segment"""

    letter = "Q"
    nargs = 4

    arg1: complex
    """The (absolute) control point"""

    arg2: complex
    """The (absolute) end point"""

    @property
    def x2(self) -> float:
        """x coordinate of the (absolute) control point"""
        return self.arg1.real

    @property
    def y2(self) -> float:
        """y coordinate of the (absolute) control point"""
        return self.arg1.imag

    @property
    def x3(self) -> float:
        """x coordinate of the (absolute) end point"""
        return self.arg2.real

    @property
    def y3(self) -> float:
        """y coordinate of the (absolute) end point"""
        return self.arg2.imag

    @property
    def args(self):
        return self.x2, self.y2, self.x3, self.y3

    @overload
    def __init__(self, x2: ComplexLike, x3: ComplexLike): ...

    @overload
    def __init__(self, x2: float, y2: float, x3: float, y3: float): ...

    def __init__(self, x2, y2, x3=None, y3=None):
        if x3 is not None:
            self.arg1 = x2 + y2 * 1j
            self.arg2 = x3 + y3 * 1j
        else:
            self.arg1, self.arg2 = complex(x2), complex(y2)

    def update_bounding_box(self, first, last_two_points, bbox):
        x1, x2, x3 = last_two_points[-1].real, self.x2, self.x3
        y1, y2, y3 = last_two_points[-1].imag, self.y2, self.y3

        if not (x1 in bbox.x and x2 in bbox.x and x3 in bbox.x):
            bbox.x += quadratic_extrema(x1, x2, x3)

        if not (y1 in bbox.y and y2 in bbox.y and y3 in bbox.y):
            bbox.y += quadratic_extrema(y1, y2, y3)

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return (self.arg1, self.arg2)

    def to_relative(self, prev: ComplexLike) -> quadratic:
        return quadratic(self.arg1 - prev, self.arg2 - prev)

    def transform(self, transform: Transform) -> Quadratic:
        return Quadratic(
            transform.capply_to_point(self.arg1), transform.capply_to_point(self.arg2)
        )

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg2

    def ccurve_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        pt1 = 1.0 / 3 * prev + 2.0 / 3 * self.arg1
        pt2 = 2.0 / 3 * self.arg1 + 1.0 / 3 * self.arg2
        return pt1, pt2, self.arg2

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Quadratic:
        prev = complex(prev)
        return Quadratic(self.x2, self.y2, prev.real, prev.imag)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Quadratic, Quadratic]:
        return self._abssplit(prev, prev_control, t)


class quadratic(QuadraticMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Relative quadratic line segment"""

    letter = "q"
    nargs = 4

    arg1: complex
    """The (relative) control point"""

    arg2: complex
    """The (relative) end point"""

    @property
    def dx2(self) -> float:
        """x coordinate of the (relative) control point"""
        return self.arg1.real

    @property
    def dy2(self) -> float:
        """y coordinate of the (relative) control point"""
        return self.arg1.imag

    @property
    def dx3(self) -> float:
        """x coordinate of the (relative) end point"""
        return self.arg2.real

    @property
    def dy3(self) -> float:
        """y coordinate of the (relative) end point"""
        return self.arg2.imag

    @property
    def args(self):
        return self.dx2, self.dy2, self.dx3, self.dy3

    @overload
    def __init__(self, dx2: ComplexLike, dx3: ComplexLike): ...

    @overload
    def __init__(self, dx2: float, dy2: float, dx3: float, dy3: float): ...

    def __init__(self, dx2, dy2, dx3=None, dy3=None):
        if dx3 is not None:
            self.arg1 = dx2 + dy2 * 1j
            self.arg2 = dx3 + dy3 * 1j
        else:
            self.arg1, self.arg2 = complex(dx2), complex(dy2)

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return (self.arg1 + prev, self.arg2 + prev)

    def to_absolute(self, prev: ComplexLike) -> Quadratic:
        return Quadratic(self.arg1 + prev, self.arg2 + prev)

    def ccurve_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        pt1 = 1.0 / 3 * prev + 2.0 / 3 * (prev + self.arg1)
        pt2 = 2.0 / 3 * (prev + self.arg1) + 1.0 / 3 * (prev + self.arg2)
        return pt1, pt2, prev + self.arg2

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg2 + prev

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> quadratic:
        return quadratic(-self.arg2 + self.arg1, -self.arg2)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[quadratic, quadratic]:
        return self._relsplit(prev, prev_control, t)


class TepidQuadratic(QuadraticMixin, AbsolutePathCommand):
    """Continued Quadratic Line segment"""

    letter = "T"
    nargs = 2

    arg1: complex
    """The (absolute) control point"""

    @property
    def x3(self) -> float:
        """x coordinate of the (absolute) end point"""
        return self.arg1.real

    @property
    def y3(self) -> float:
        """y coordinate of the (absolute) end point"""
        return self.arg1.imag

    @property
    def args(self):
        return self.x3, self.y3

    @overload
    def __init__(self, x3: ComplexLike): ...

    @overload
    def __init__(self, x3: float, y3: float): ...

    def __init__(self, x3, y3=None):
        if y3 is not None:
            self.arg1 = x3 + y3 * 1j
        else:
            self.arg1 = complex(x3)

    def update_bounding_box(self, first, last_two_points, bbox):
        self.to_quadratic(last_two_points[-1], last_two_points[-2]).update_bounding_box(
            first, last_two_points, bbox
        )

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return (2 * prev - prev_prev, self.arg1)

    def to_non_shorthand(
        self, prev: ComplexLike, prev_control: ComplexLike
    ) -> Quadratic:
        return self.to_quadratic(prev, prev_control)

    def to_relative(self, prev: ComplexLike) -> tepidQuadratic:
        return tepidQuadratic(self.arg1 - prev)

    def transform(self, transform: Transform) -> TepidQuadratic:
        return TepidQuadratic(transform.capply_to_point(self.arg1))

    def ccurve_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        qp1 = 2 * prev - prev_prev
        qp2 = self.arg1
        pt1 = 1.0 / 3 * prev + 2.0 / 3 * qp1
        pt2 = 2.0 / 3 * qp1 + 1.0 / 3 * qp2
        return pt1, pt2, qp2

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg1

    def to_quadratic(self, prev: ComplexLike, prev_prev: ComplexLike) -> Quadratic:
        """Convert this continued quadratic into a full quadratic"""
        return Quadratic(
            *self.ccontrol_points(complex(prev), complex(prev), complex(prev_prev))
        )

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> TepidQuadratic:
        return TepidQuadratic(prev)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Quadratic, Quadratic]:
        return self._abssplit(prev, prev_control, t)


class tepidQuadratic(QuadraticMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Relative continued quadratic line segment"""

    letter = "t"
    nargs = 2

    arg1: complex
    """The (relative) control point"""

    @property
    def dx3(self) -> float:
        """x coordinate of the (relative) end point"""
        return self.arg1.real

    @property
    def dy3(self) -> float:
        """y coordinate of the (relative) end point"""
        return self.arg1.imag

    @property
    def args(self):
        return self.dx3, self.dy3

    @overload
    def __init__(self, dx3: ComplexLike): ...

    @overload
    def __init__(self, dx3: float, dy3: float): ...

    def __init__(self, dx3, dy3=None):
        if dy3 is not None:
            self.arg1 = dx3 + dy3 * 1j
        else:
            self.arg1 = complex(dx3)

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return (2 * prev - prev_prev, self.arg1 + prev)

    def ccurve_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        qp1 = 2 * prev - prev_prev
        qp2 = self.arg1 + prev
        pt1 = 1.0 / 3 * prev + 2.0 / 3 * qp1
        pt2 = 2.0 / 3 * qp1 + 1.0 / 3 * qp2
        return pt1, pt2, qp2

    def to_absolute(self, prev: ComplexLike) -> TepidQuadratic:
        return TepidQuadratic(self.arg1 + prev)

    def to_non_shorthand(
        self, prev: ComplexLike, prev_control: ComplexLike
    ) -> Quadratic:
        return self.to_absolute(prev).to_non_shorthand(prev, prev_control)

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg1 + prev

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> tepidQuadratic:
        return tepidQuadratic(-self.arg1)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[quadratic, quadratic]:
        return self._relsplit(prev, prev_control, t)



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (c) 2024 jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Convert CGM files to SVG
"""

from dataclasses import dataclass
import inspect
from inkex.utils import pairwise
import math
from typing import Callable, Dict, List, Optional, Tuple, get_type_hints, Union

import inkex
import cgm_parse
import cgm_enums


class CGMConverterMeta(type):
    """Metaclass so that we can automatically call the methods based
    on their type hint"""

    def __new__(cls, name, bases, class_dict):
        new_class = super().__new__(cls, name, bases, class_dict)

        call_mapping = {}
        for name, method in inspect.getmembers(new_class, predicate=inspect.isfunction):
            hints = get_type_hints(method)
            if name.startswith("on_") and len(hints) == 1:
                type_ = list(hints.values())[0]

                call_mapping[type_] = method

        new_class.call_mapping = call_mapping
        return new_class


@dataclass
class FillAttributes:
    fill: cgm_enums.Colour
    interior_style: cgm_enums.InteriorStyleEnum


@dataclass
class LineAttributes:
    specification_mode: cgm_enums.WidthSpecificationModeEnum
    width: float
    colour: cgm_enums.Colour
    type: cgm_enums.LineTypeEnum
    visibility: cgm_enums.EdgeVisibilityEnum


@dataclass
class TextAttributes:
    colour: cgm_enums.Colour
    character_height: Optional[float]
    fallback_character_height: Optional[float]
    horizontal_alignment: cgm_enums.TextHorizontalAlignmentEnum
    vertical_alignment: cgm_enums.TextVerticalAlignmentEnum
    x_character_up: Optional[float]
    y_character_up: Optional[float]
    x_character_base: Optional[float]
    y_character_base: Optional[float]
    precision: cgm_enums.TextPrecisionEnum
    expansion_ratio: float
    character_spacing: float
    path: cgm_enums.TextPathEnum
    font_index: int


class CGMConverter(metaclass=CGMConverterMeta):
    call_mapping: Dict[type, Callable]

    def __init__(self) -> None:
        self.current_page = None
        self.document = inkex.InputExtension.get_template(width="0", height="0")
        self.svg = self.document.getroot()
        self.color_table: Dict[int, cgm_enums.DirectColour] = {}
        self.vdc_height = 100
        self.font_list: List[str] = ["sans-serif"]
        self.default_replacement: List[cgm_parse.CGMCommand] = []
        self.first_page_origin: Optional[Tuple[float, float]] = None

        pass

    def set_defaults(self) -> None:
        self.edge_attributes = LineAttributes(
            cgm_enums.WidthSpecificationModeEnum.SCALED,
            1.0,
            cgm_enums.DirectColour(0, 0, 0),  # device-dependent fg color
            cgm_enums.LineTypeEnum.SOLID,
            cgm_enums.EdgeVisibilityEnum.OFF,
        )
        self.line_attributes = LineAttributes(
            cgm_enums.WidthSpecificationModeEnum.SCALED,
            1.0,
            cgm_enums.DirectColour(0, 0, 0),  # device-dependent fg color
            cgm_enums.LineTypeEnum.SOLID,
            cgm_enums.EdgeVisibilityEnum.ON,
        )
        self.fill_attributes = FillAttributes(
            cgm_enums.DirectColour(0, 0, 0),  # device-dependent fg color
            interior_style=cgm_enums.InteriorStyleEnum.HOLLOW,
        )
        self.text_attributes = TextAttributes(
            cgm_enums.DirectColour(0, 0, 0),  # device-dependent fg color
            None,
            None,
            cgm_enums.TextHorizontalAlignmentEnum.NORMAL_HORIZONTAL,
            cgm_enums.TextVerticalAlignmentEnum.NORMAL_VERTICAL,
            0,
            None,
            None,
            0,
            cgm_enums.TextPrecisionEnum.STRING,
            1.0,
            0,
            cgm_enums.TextPathEnum.RIGHT,
            1,
        )
        self.current_text: Optional[inkex.TextElement] = None
        first = self.first_page_origin == None
        for cmd in self.default_replacement:
            self.stream(cmd)
        if first:
            # Restore this information if it was overwritten by the default replacement
            self.first_page_origin = None

    def apply_edge(self, elem):
        if self.edge_attributes.visibility == cgm_enums.EdgeVisibilityEnum.ON:
            elem.style.update(
                {
                    "stroke": self.get_colour(self.edge_attributes.colour),
                    "stroke-width": self.edge_attributes.width,
                }
            )
            self.apply_stroke_dasharray(elem, self.edge_attributes.type)

    def apply_line(self, elem):
        elem.style.update(
            {
                "fill": None,
                "stroke": self.get_colour(self.line_attributes.colour),
                "stroke-width": self.line_attributes.width,
            }
        )
        self.apply_stroke_dasharray(elem, self.line_attributes.type)

    def apply_stroke_dasharray(self, elem, stroke_type: cgm_enums.LineTypeEnum):
        if stroke_type == cgm_enums.LineTypeEnum.SOLID:
            return
        elif stroke_type == cgm_enums.LineTypeEnum.DASH:
            array = [3, 3]
        elif stroke_type == cgm_enums.LineTypeEnum.DOT:
            array = [1, 1]
        elif stroke_type == cgm_enums.LineTypeEnum.DASH_DOT:
            array = [4, 2, 1, 2]
        elif stroke_type == cgm_enums.LineTypeEnum.DASH_DOT_DOT:
            array = [4, 2, 1, 2, 1, 2]
        else:
            return  # private value
        elem.style["stroke-dasharray"] = [
            i * float(elem.style["stroke-width"]) for i in array
        ]

    def get_colour(self, colour: cgm_enums.Colour):
        if isinstance(colour, cgm_enums.IndexColour):
            colour = self.color_table.get(
                colour.index,
                cgm_enums.DirectColour(255, 255, 255)
                if colour.index == 0
                else cgm_enums.DirectColour(0, 0, 0),
            )
        if isinstance(colour, cgm_enums.DirectColour):
            if colour.colour_model == cgm_enums.ColourModelEnum.RGB:
                return inkex.Color([colour.v1, colour.v2, colour.v3])
            else:
                inkex.errormsg(f"Color model {colour.colour_model} not implemented")

    def apply_fill(self, elem):
        fill = None
        if self.fill_attributes.interior_style == cgm_enums.InteriorStyleEnum.HOLLOW:
            fill = None
        elif self.fill_attributes.interior_style == cgm_enums.InteriorStyleEnum.SOLID:
            fill = self.get_colour(self.fill_attributes.fill)
        elem.style.update({"fill-rule": "evenodd", "fill": fill})

    def on_begin_picture(self, arg: cgm_parse.BeginPicture):
        self.current_page = self.svg.namedview.add(inkex.Page())
        assert self.current_page is not None
        self.current_page.set("inkscape:label", str(arg.name))
        self.current_layer: inkex.Layer = self.svg.add(inkex.Layer())

        self.set_defaults()

    def on_vdc_extent(self, arg: cgm_parse.VDCExtent):
        first_page = False
        if self.first_page_origin is None:
            self.first_page_origin = arg.first_corner.x, arg.first_corner.y
            first_page = True

        x, y, width, height = (
            arg.first_corner.x - self.first_page_origin[0],
            arg.first_corner.y - self.first_page_origin[1],
            abs(arg.second_corner.x - arg.first_corner.x),
            abs(arg.second_corner.y - arg.first_corner.y),
        )
        assert self.current_page is not None

        self.current_page.set("x", x)
        self.current_page.set("y", y)
        self.current_page.set("width", width)
        self.current_page.set("height", height)
        self.svg.set("width", width)
        self.svg.set("height", height)
        # x1 = x * matrix.a + y * matrix.b + matrix.e,
        # y1 = x * matrix.c + y * matrix.d + matrix.f;

        # We need a transform so that VDC Extent  p1 is the bottom left of the page
        # and VDC Extent p2 is the bottom right of the page

        a = (width) / (arg.second_corner.x - arg.first_corner.x)
        d = (height) / (arg.first_corner.y - arg.second_corner.y)
        e = x - arg.first_corner.x * a
        f = y - arg.second_corner.y * d

        trans = inkex.Transform((a, 0, 0, d, e, f))
        assert (
            trans.apply_to_point(arg.first_corner.x + arg.first_corner.y * 1j)
            == x + (y + height) * 1j
        )
        assert (
            trans.apply_to_point(arg.second_corner.x + arg.second_corner.y * 1j)
            == (x + width) + (y) * 1j
        )

        self.current_layer.transform = trans
        if first_page:
            # only for first page
            self.svg.set("viewBox", f"0 0 {width} {height}")
        self.vdc_height = height
        pass

    def on_rectangle(self, arg: cgm_parse.Rectangle):
        rect = self.current_layer.add(
            inkex.Rectangle.new(
                arg.p1.x, arg.p1.y, arg.p2.x - arg.p1.x, arg.p2.y - arg.p1.y
            )
        )
        self.apply_fill(rect)
        self.apply_edge(rect)

    def on_circle(self, arg: cgm_parse.Circle):
        circle = self.current_layer.add(
            inkex.Circle.new((arg.center.x, arg.center.y), arg.radius)
        )
        self.apply_fill(circle)
        self.apply_edge(circle)

    def on_ellipse(self, arg: cgm_parse.Ellipse):
        ellipse = inkex.Ellipse.new((0, 0), (1, 1))

        ellipse.transform = inkex.Transform(
            (
                arg.point_1.x - arg.center.x,
                arg.point_1.y - arg.center.y,
                arg.point_2.x - arg.center.x,
                arg.point_2.y - arg.center.y,
                arg.center.x,
                arg.center.y,
            )
        )
        assert (
            ellipse.transform.apply_to_point(1) == arg.point_1.x + (arg.point_1.y) * 1j
        )
        assert (
            ellipse.transform.apply_to_point(1j) == arg.point_2.x + (arg.point_2.y) * 1j
        )
        pel: inkex.PathElement = self.current_layer.add(ellipse.to_path_element())
        pel.apply_transform()
        self.apply_fill(pel)
        self.apply_edge(pel)

    def on_polygon(self, arg: cgm_parse.Polygon):
        poly = inkex.Polygon.new(" ".join([f"{i.x},{i.y}" for i in arg.points]))

        path = self.current_layer.add(poly.to_path_element())
        self.apply_fill(path)
        self.apply_edge(path)

    def on_polyline(self, arg: cgm_parse.Polyline):
        poly = inkex.Polyline.new(" ".join([f"{i.x},{i.y}" for i in arg.points]))

        path = self.current_layer.add(poly.to_path_element())
        self.apply_line(path)

    def on_disjoint_polyline(self, arg: cgm_parse.DisjointPolyline):
        path = inkex.PathElement.new(
            inkex.Path(
                [
                    inkex.paths.Move(pt.x, pt.y)
                    if i % 2 == 0
                    else inkex.paths.Line(pt.x, pt.y)
                    for i, pt in enumerate(arg.points)
                ]
            )
        )
        self.current_layer.add(path)
        self.apply_line(path)

    def on_polygon_set(self, arg: cgm_parse.PolygonSet):
        current_close_point = cgm_enums.Point(0, 0)
        if len(arg.points) == 0:
            return

        polygons: List[List[Tuple[bool, cgm_enums.Point]]] = []

        # First construct a list of polygons in the set.
        def iterator(pts):
            yield from pairwise(pts)
            yield [pts[-1], None]

        for e1, e2 in iterator(arg.points):
            if e1 is None:
                polygons.append([])
                current_close_point = e2.pt
            else:
                if e2 is not None and e1.linetype in (
                    cgm_enums.PolygonSetLineTypeEnum.INVISIBLE,
                    cgm_enums.PolygonSetLineTypeEnum.VISIBLE,
                ):
                    polygons[-1].append(
                        (e1.linetype == cgm_enums.PolygonSetLineTypeEnum.VISIBLE, e2.pt)
                    )
                else:
                    draw = e1.linetype in (
                        cgm_enums.PolygonSetLineTypeEnum.CLOSE_VISIBLE,
                        cgm_enums.PolygonSetLineTypeEnum.VISIBLE,
                    )
                    polygons[-1].append((draw, current_close_point))
                    if e2 is not None:
                        current_close_point = e2.pt
                        polygons.append([])

        # Now re-shuffle each polygon so we start with an invisible
        # line if it exists. Coordinates are absolute so this should be fine
        shuffled = []
        single_element = True
        for poly in polygons:
            has_invisible = any([not j[0] for j in poly])
            if has_invisible:
                inv_index = ([p[0] for p in poly]).index(False)
                shuffled.append(poly[inv_index:] + poly[:inv_index])
            else:
                shuffled.append(poly)
            if has_invisible:
                single_element = False

        # Draw the fill, then the stroke.
        fill_path = inkex.Path()
        for poly in shuffled:
            fill_path.append(inkex.paths.Move(poly[-1][1].x, poly[-1][1].y))
            for point in poly[:-1]:  # The last line is drawn via ZoneClose
                fill_path.append(inkex.paths.Line(point[1].x, point[1].y))
            fill_path.append(inkex.paths.ZoneClose())
        fill_elem = self.current_layer.add(inkex.PathElement.new(fill_path))
        self.apply_fill(fill_elem)

        if single_element:
            self.apply_edge(fill_elem)

        else:
            # Draw the stroke separately
            stroke_path = inkex.Path()

            for poly in shuffled:
                stroke_path.append(inkex.paths.Move(poly[-1][1].x, poly[-1][1].y))
                for point in poly:
                    if point[0]:
                        stroke_path.append(inkex.paths.Line(point[1].x, point[1].y))
                    else:
                        stroke_path.append(inkex.paths.Move(point[1].x, point[1].y))
                if all(i[0] for i in poly):
                    stroke_path.append(inkex.paths.ZoneClose())
            stroke_elem = self.current_layer.add(inkex.PathElement.new(stroke_path))
            self.apply_edge(stroke_elem)
            stroke_elem.style["fill"] = None

    def on_polybezier(self, arg: cgm_parse.Polybezier):
        if len(arg.points) == 0:
            return

        path = inkex.Path()
        if arg.continuous == cgm_enums.PolybezierContinuityEnum.CONTINUOUS:
            assert len(arg.points) % 3 == 1, "Bad Polybezier specification"
            path.append(inkex.paths.Move(complex(arg.points[0])))
            for i in range((len(arg.points) - 1) // 3):
                path.append(
                    inkex.paths.Curve(
                        complex(arg.points[3 * i + 1]),
                        complex(arg.points[3 * i + 2]),
                        complex(arg.points[3 * i + 3]),
                    )
                )
        else:
            assert len(arg.points) % 4 == 0, "Bad Polybezier specification"
            for i in range(len(arg.points) // 4):
                if i == 0 or path[-1].cend_point(0, 0) != complex(arg.points[4 * i]):
                    path.append(inkex.paths.Move(complex(arg.points[4 * i])))
                path.append(
                    inkex.paths.Curve(
                        complex(arg.points[4 * i + 1]),
                        complex(arg.points[4 * i + 2]),
                        complex(arg.points[4 * i + 3]),
                    )
                )
        pel = self.current_layer.add(inkex.PathElement.new(path))
        self.apply_line(pel)
        pel.style["stroke"] = "red"

    def on_circular_arc_3pt(self, arg: cgm_parse.CircularArc3Point):
        path = inkex.PathElement.arc_from_3_points(
            complex(arg.p1), complex(arg.p2), complex(arg.p3), "arc"
        )
        self.current_layer.append(path)
        self.apply_line(path)

    def on_circular_arc_3pt_close(self, arg: cgm_parse.CircularArc3PointClose):
        path = inkex.PathElement.arc_from_3_points(
            complex(arg.p1),
            complex(arg.p2),
            complex(arg.p3),
            "chord" if arg.closure == cgm_enums.ArcClosureEnum.CHORD else "slice",
        )
        self.current_layer.append(path)
        self.apply_fill(path)
        self.apply_edge(path)

    def circular_arc_helper(
        self, arg: Union[cgm_parse.CircularArcCentre, cgm_parse.CircularArcCentreClose]
    ):
        c = complex(arg.centre)
        start_ray = arg.delta_x_start + arg.delta_y_start * 1j
        end_ray = arg.delta_x_end + arg.delta_y_end * 1j
        start_angle = inkex.Vector2d(start_ray).angle
        end_angle = inkex.Vector2d(end_ray).angle
        arctype = "arc"
        if isinstance(arg, cgm_parse.CircularArcCentreClose):
            arctype = (
                "chord" if arg.closure == cgm_enums.ArcClosureEnum.CHORD else "slice"
            )
        path = inkex.PathElement.arc(
            inkex.Vector2d(c),
            arg.radius,
            arg.radius,
            arctype,
            start=start_angle,
            end=end_angle,
        )
        return path

    def on_circular_arc_centre(self, arg: cgm_parse.CircularArcCentre):
        path = self.circular_arc_helper(arg)
        self.current_layer.append(path)
        self.apply_line(path)

    def on_circular_arc_centre_close(self, arg: cgm_parse.CircularArcCentreClose):
        path = self.circular_arc_helper(arg)
        self.current_layer.append(path)
        self.apply_fill(path)
        self.apply_edge(path)

    def on_interior_style(self, arg: cgm_parse.InteriorStyle):
        self.fill_attributes.interior_style = arg.interior_style

    def on_fill_colour(self, arg: cgm_parse.FillColour):
        self.fill_attributes.fill = arg.fill

    def on_color_table(self, arg: cgm_parse.ColourTable):
        for i, col in enumerate(arg.colours):
            self.color_table[i + arg.colour_index.index] = col

    def on_line_width(self, arg: cgm_parse.LineWidth):
        if arg.width > 0:
            self.line_attributes.width = arg.width

    def on_line_width_specification_mode(
        self, arg: cgm_parse.LineWidthSpecificationMode
    ):
        self.line_attributes.specification_mode = arg.size_specification

    def on_line_colour(self, arg: cgm_parse.LineColour):
        self.line_attributes.colour = arg.colour

    def on_line_type(self, arg: cgm_parse.LineType):
        self.line_attributes.type = arg.type

    def on_edge_width(self, arg: cgm_parse.EdgeWidth):
        if arg.width > 0:
            self.edge_attributes.width = arg.width

    def on_edge_colour(self, arg: cgm_parse.EdgeColour):
        self.edge_attributes.colour = arg.colour

    def on_edge_type(self, arg: cgm_parse.EdgeType):
        self.edge_attributes.type = arg.type

    def on_edge_visibility(self, arg: cgm_parse.EdgeVisibility):
        self.edge_attributes.visibility = arg.visibility

    def on_background(self, arg: cgm_parse.BackgroundColour):
        self.svg.namedview.set("pagecolor", str(self.get_colour(arg.colour)))

    # Text-related

    def on_font_list(self, arg: cgm_parse.FontList):
        self.font_list = arg.fonts

    def on_text_font_index(self, arg: cgm_parse.TextFontIndex):
        self.text_attributes.font_index = arg.index

    def on_text_alignment(self, arg: cgm_parse.TextAlignment):
        self.text_attributes.horizontal_alignment = arg.horizontal_alignment
        self.text_attributes.vertical_alignment = arg.vertical_alignment

    def on_character_height(self, arg: cgm_parse.CharacterHeight):
        self.text_attributes.character_height = arg.height

    def on_character_orientation(self, arg: cgm_parse.CharacterOrientation):
        self.text_attributes.x_character_base = arg.x_character_base
        self.text_attributes.x_character_up = arg.x_character_up
        self.text_attributes.y_character_base = arg.y_character_base
        self.text_attributes.y_character_up = arg.y_character_up
        self.text_attributes.fallback_character_height = math.sqrt(
            arg.x_character_up**2 + arg.y_character_up**2
        )

    def on_text_precision(self, arg: cgm_parse.TextPrecision):
        self.text_attributes.precision = arg.text_precision

    def on_character_spacing(self, arg: cgm_parse.CharacterSpacing):
        self.text_attributes.character_spacing = arg.spacing

    def on_text_path(self, arg: cgm_parse.TextPath):
        self.text_attributes.path = arg.path

    def on_character_expansion(self, arg: cgm_parse.CharacterExpansionFactor):
        self.text_attributes.expansion_ratio = arg.factor

    def on_text_colour(self, arg: cgm_parse.TextColour):
        self.text_attributes.colour = arg.colour

    def add_tspan(self, text: str):
        assert self.current_text is not None
        tspan = self.current_text.add(inkex.Tspan())
        tspan.text = text
        ch = (
            self.text_attributes.character_height
            or self.text_attributes.fallback_character_height
        )
        tspan.style.update(
            {
                "font-size": 1 / 100 * self.vdc_height if ch is None else ch,
                "font-stretch": f"{self.text_attributes.expansion_ratio * 100}%",
                "letter-spacing": self.text_attributes.character_spacing,
                "fill": self.get_colour(self.text_attributes.colour),
                # 1-based indices
                "font-family": self.font_list[self.text_attributes.font_index - 1],
            }
        )

    def on_text(self, arg: cgm_parse.Text):
        self.current_text = inkex.TextElement()
        # Apply transform
        # get baseline unit vector
        cuv = inkex.Vector2d(
            0
            if self.text_attributes.x_character_up is None
            else self.text_attributes.x_character_up,
            self.vdc_height
            if self.text_attributes.y_character_up is None
            else self.text_attributes.y_character_up,
        )
        cbv = inkex.Vector2d(
            self.vdc_height
            if self.text_attributes.x_character_base is None
            else self.text_attributes.x_character_base,
            0
            if self.text_attributes.y_character_base is None
            else self.text_attributes.y_character_base,
        )
        cuv = 1 / cuv.length * cuv
        cbv = 1 / cbv.length * cbv

        self.current_text.transform = inkex.Transform(
            (cbv.x, cbv.y, cuv.x, cuv.y, 0, 0)
        ) @ inkex.Transform((1, 0, 0, -1, 0, 0))
        origin = arg.point.x + arg.point.y * 1j
        # Apply reverse transform to origin
        origin = (-self.current_text.transform).apply_to_point(origin)
        self.current_text.set("x", origin.x)
        self.current_text.set("y", origin.y)

        if (
            self.text_attributes.horizontal_alignment
            == cgm_enums.TextHorizontalAlignmentEnum.CENTER
        ):
            self.current_text.style["text-anchor"] = "middle"
        elif (
            self.text_attributes.horizontal_alignment
            == cgm_enums.TextHorizontalAlignmentEnum.RIGHT
        ):
            self.current_text.style["text-anchor"] = "end"

        if (
            self.text_attributes.vertical_alignment
            == cgm_enums.TextVerticalAlignmentEnum.HALF
        ):
            self.current_text.set("dy", "0.4em")
        if (
            self.text_attributes.vertical_alignment
            == cgm_enums.TextVerticalAlignmentEnum.TOP
        ):
            self.current_text.set("dy", "1em")

        # TODO: character direction, vertical alignment

        self.add_tspan(arg.string)
        if arg.flag == cgm_enums.TextFinalFlag.FINAL:
            self.finalize_text()

    def finalize_text(self):
        max_font_size = max([i.style["font-size"] for i in self.current_text])
        # This is for vertical alignment, which is specified in em of the largest
        # text of the box
        self.current_text.style["font-size"] = max_font_size

        self.current_layer.append(self.current_text)

    def on_default_replacement(self, arg: cgm_parse.MetafileDefaultsReplacement):
        self.default_replacement = arg.data

    def stream(self, arg: cgm_parse.CGMCommand):
        handler = self.call_mapping.get(type(arg))
        if handler:  # otherwise we ignore the command
            handler(self, arg)


class CgmInput(inkex.InputExtension):
    def __init__(self) -> None:
        super().__init__()

        self.converter = CGMConverter()

    def load(self, stream):
        self.parser = cgm_parse.BinaryCGMCommandParser(stream)
        for command in self.parser.parse():
            self.converter.stream(command)

        return self.converter.document


if __name__ == "__main__":
    CgmInput().run()



"""Check .inx file(s).

This is meant to be run an executable module, e.g.::

    python -m inkex.tester.test_inx_file *.inx

"""

import argparse
import sys
import unittest
from typing import List

from .inx import InxMixin


class InxTestCase(InxMixin, unittest.TestCase):
    inx_files: List[str] = []

    def test_inx_file_parameters(self):
        for inx_file in self.inx_files:
            with self.subTest(inx_file=inx_file):
                self.assertInxIsGood(inx_file)

    def test_inx_file_schema(self):
        for inx_file in self.inx_files:
            with self.subTest(inx_file=inx_file):
                self.assertInxSchemaValid(inx_file)


def main(args=None):
    """Check .inx file(s)"""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="store_const",
        const=2,
        help="Verbose output",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="verbosity",
        action="store_const",
        const=0,
        help="Quiet output",
    )
    parser.add_argument(
        "inx_files", metavar="INX-FILE", nargs="+", help="An .inx file to check"
    )
    args = parser.parse_args(args, namespace=argparse.Namespace(verbosity=1))

    InxTestCase.inx_files = args.inx_files
    runner = unittest.TextTestRunner(verbosity=args.verbosity)

    suite = unittest.defaultTestLoader.loadTestsFromTestCase(InxTestCase)
    result = runner.run(suite)
    sys.exit(not result.wasSuccessful())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright (C) 2009 Aurelio A. Heckert, aurium (a) gmail dot com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This effect adds a feature visible (or usable) only on a SVG enabled web
browser (like Firefox). This effect sets one or more attributes in the second
selected element, when a defined event occurs on the first selected element.
If you want to set more than one attribute, you must separate this with a space,
and only with aspace.
"""

# local library
import inkex
from inkex.localization import inkex_gettext as _
import inkwebeffect


class SetAttribute(inkwebeffect.InkWebEffect):
    """Set a web attribute accross many objects"""

    def add_arguments(self, pars):
        pars.add_argument("--tab", help="The selected UI-tab when OK was pressed")
        pars.add_argument(
            "--att", default="fill stroke stroke-width", help="Attribute to set."
        )
        pars.add_argument("--val", default="red black 5px", help="Values to set.")
        pars.add_argument("--when", default="onclick", help="When it must to set?")
        pars.add_argument("--from-and-to", dest="from_and_to", default="g-to-one")
        pars.add_argument(
            "--compatibility",
            default="append",
            help="Compatibility with previews code to this event.",
        )

    def effect(self):
        self.ensureInkWebSupport()

        if len(self.options.ids) < 2:
            raise inkex.AbortExtension(_("You must select at least two elements."))

        # All set the last else The first set all
        split = -1 if self.options.from_and_to == "g-to-one" else 1
        el_from = list(self.svg.selection.values())[:split]
        id_to = list(self.svg.selection.ids)[split:]

        ev_code = "InkWeb.setAtt({{el:['{}'], att:'{}', val:'{}'}})".format(
            "','".join(id_to), self.options.att, self.options.val
        )
        for elem in el_from:
            prev_ev_code = elem.get(self.options.when)
            if prev_ev_code is None:
                prev_ev_code = ""

            if self.options.compatibility == "append":
                el_ev_code = prev_ev_code + ";\n" + ev_code
            if self.options.compatibility == "prepend":
                el_ev_code = ev_code + ";\n" + prev_ev_code
            if self.options.compatibility == "replace":
                el_ev_code = ev_code

            elem.set(self.options.when, el_ev_code)


if __name__ == "__main__":
    SetAttribute().run()



#!/usr/bin/env python3
"""Remove colors"""

import inkex


class Desaturate(inkex.ColorExtension):
    """Remove color but maintain intesity"""

    def modify_color(self, name, color):
        lum = (
            max(color.red, color.green, color.blue)
            + min(color.red, color.green, color.blue)
        ) // 2
        return inkex.Color((int(round(lum)), int(round(lum)), int(round(lum))))


if __name__ == "__main__":
    Desaturate().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2019 Marc Jeanmougin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
"""
Generate Latex via a PDF using pdflatex
"""

import os

import inkex
from inkex.base import TempDirMixin
from inkex.command import ProgramRunError, call, inkscape
from inkex import load_svg, ShapeElement, Defs
from inkex.units import convert_unit
from inkex.localization import inkex_gettext as _


class PdfLatex(TempDirMixin, inkex.GenerateExtension):
    """
    Use pdflatex to generate LaTeX, this whole hack is required because
    we don't want to open a LaTeX document as a document, but as a
    generated fragment (like import, but done manually).
    """

    def add_arguments(self, pars):
        pars.add_argument(
            "--formule",
            type=str,
            default=r"\(\displaystyle\frac{\pi^2}{6}=\lim_{n \to \infty}\sum_{k=1}^n \frac{1}{k^2}\)",
        )
        pars.add_argument(
            "--preamble",
            type=str,
            default="".join(
                f"\\usepackage{{ams{i}}}" for i in ["math", "symb", "fonts"]
            ),
        )
        pars.add_argument("--font_size", type=int, default=10)
        pars.add_argument("--page", choices=["basic", "advanced"])
        pars.add_argument("--standalone", type=inkex.Boolean, default=True)

    def generate(self):
        tex_file = os.path.join(self.tempdir, "input.tex")
        pdf_file = os.path.join(self.tempdir, "input.pdf")  # Auto-generate by pdflatex
        svg_file = os.path.join(self.tempdir, "output.svg")

        with open(tex_file, "w") as fhl:
            self.write_latex(fhl)

        try:
            call(
                "pdflatex",
                tex_file,
                output_directory=self.tempdir,
                halt_on_error=True,
                oldie=True,
            )
        except ProgramRunError as err:
            inkex.errormsg(_("An exception occurred during LaTeX compilation: ") + "\n")
            inkex.errormsg(
                err.stdout.decode("utf8")
                .replace("\r\n", "\n")
                .split("Transcript written on")[0]
            )
            raise inkex.AbortExtension()

        inkscape(
            pdf_file,
            export_filename=svg_file,
            pages=1,
            pdf_poppler=True,
            export_type="svg",
            actions="select-all;page-fit-to-selection;clone-unlink-recursively;vacuum-defs",
        )

        if not os.path.isfile(svg_file):
            fn = os.path.basename(svg_file)
            if os.path.isfile(fn):
                # Inkscape bug detected, file got saved wrong
                svg_file = fn

        with open(svg_file, "r") as fhl:
            svg: inkex.SvgDocumentElement = load_svg(fhl).getroot()

            scale = convert_unit("1pt", "px") / self.svg.scale
            if not self.options.standalone:
                scale *= self.options.font_size / 10

            for child in svg:
                if isinstance(child, ShapeElement):
                    child.transform = inkex.Transform(scale=scale) @ child.transform
                    yield child
                elif isinstance(child, Defs):
                    for def_child in child:
                        # The ids of both the imported and the target document should
                        # be off-limits
                        def_child.set_random_ids(
                            backlinks=True, blacklist=self.svg.get_ids()
                        )
                        self.svg.defs.append(def_child)

    def write_latex(self, stream):
        """Takes a formula and wraps it in latex"""
        if self.options.standalone:
            docclass = (
                f"\\documentclass[fontsize={self.options.font_size}pt, "
                + "class=scrreprt, preview, border=2pt]{standalone}"
            )
        else:
            docclass = r"\documentclass{minimal}"
        stream.write(
            f"""%% processed with pdflatex.py
{docclass}
{self.options.preamble}
\\begin{{document}}
{self.options.formule}
\\end{{document}}
"""
        )
        stream.flush()


if __name__ == "__main__":
    PdfLatex().run()



# coding=utf-8
#
# Copyright (C) 2007 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Python barcode renderer for RM4CC barcodes. Designed for use with Inkscape.
"""

from .Base import Barcode

charmap = {
    "(": "25",
    ")": "3",
    "0": "05053535",
    "1": "05152535",
    "2": "05153525",
    "3": "15052535",
    "4": "15053525",
    "5": "15152525",
    "6": "05251535",
    "7": "05350535",
    "8": "05351525",
    "9": "15250535",
    "A": "15251525",
    "B": "15350525",
    "C": "05253515",
    "D": "05352515",
    "E": "05353505",
    "F": "15252515",
    "G": "15253505",
    "H": "15352505",
    "I": "25051535",
    "J": "25150535",
    "K": "25151525",
    "L": "35050535",
    "M": "35051525",
    "N": "35150525",
    "O": "25053525",
    "P": "25152515",
    "Q": "25153505",
    "R": "35052515",
    "S": "35053505",
    "T": "35152505",
    "U": "25251515",
    "V": "25350515",
    "W": "25351505",
    "X": "35250515",
    "Y": "35251505",
    "Z": "35350505",
}

check = ["ZUVWXY", "501234", "B6789A", "HCDEFG", "NIJKLM", "TOPQRS"]
(BAR_TRACK, BAR_DOWN, BAR_UP, BAR_FULL, BAR_NONE, WHITE_SPACE) = range(6)


class Rm4scc(Barcode):
    """Provice a Rm4scc barcode generator"""

    default_height = 18

    def encode(self, text):
        result = ""

        text = text.upper()
        text.replace("(", "")
        text.replace(")", "")

        text = "(" + text + Rm4scc.checksum(text) + ")"

        i = 0
        for char in text:
            if char in charmap:
                result = result + charmap[char]
                i += 1

        return result

    @staticmethod
    def checksum(text):
        """given a string of data, return the check character"""
        total_lower = 0
        total_upper = 0
        for char in text:
            if char in charmap:
                bars = charmap[char][0:8:2]
                lower = 0
                upper = 0

                if int(bars[0]) & 1:
                    lower += 4
                if int(bars[1]) & 1:
                    lower += 2
                if int(bars[2]) & 1:
                    lower += 1
                if int(bars[0]) & 2:
                    upper += 4
                if int(bars[1]) & 2:
                    upper += 2
                if int(bars[2]) & 2:
                    upper += 1
            total_lower += lower % 6
            total_upper += upper % 6

        total_lower = total_upper % 6
        total_upper %= 6

        checkchar = check[total_upper][total_lower]
        return checkchar

    def get_style(self, index):
        """Royal Mail Barcodes use a completely different style"""
        result = {"width": 2, "write": True, "top": 0}
        if index == BAR_TRACK:  # Track Bar
            result["top"] = 6
            result["height"] = 5
        elif index == BAR_DOWN:  # Decender Bar
            result["top"] = 6
            result["height"] = 11
        elif index == BAR_UP:  # Accender Bar
            result["height"] = 11
        elif index == BAR_FULL:  # Full Bar
            result["height"] = 17
        elif index == WHITE_SPACE:  # White Space
            result["write"] = False
        return result



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Carsten Goetze c.goetze@tu-bs.de
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import math
import random
import inkex
from inkex.paths import Move, Line


def calculate_subdivision(smoothness, x1, y1, x2, y2):
    # Calculate the vector from (x1,y1) to (x2,y2)
    x3 = x2 - x1
    y3 = y2 - y1
    # Calculate the point half-way between the two points
    hx = x1 + x3 / 2
    hy = y1 + y3 / 2
    # Calculate normalized vector perpendicular to the vector (x3,y3)
    length = math.sqrt(x3 * x3 + y3 * y3)
    if length != 0:
        nx = -y3 / length
        ny = x3 / length
    else:
        nx = 1
        ny = 0
    # Scale perpendicular vector by random factor """
    r = random.uniform(-length / (1 + smoothness), length / (1 + smoothness))
    nx = nx * r
    ny = ny * r
    # add scaled perpendicular vector to the half-way point to get the final displaced subdivision point
    x = hx + nx
    y = hy + ny
    return (x, y)


class Fractalize(inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument(
            "-s", "--subdivs", type=int, default=6, help="Number of subdivisons"
        )
        pars.add_argument(
            "-f",
            "--smooth",
            type=float,
            default=4.0,
            help="Smoothness of the subdivision",
        )

    def effect(self):
        for node in self.svg.selection.filter(inkex.PathElement):
            path = node.path.to_absolute()
            result = []
            for cmd_proxy in path.proxy_iterator():  # type: inkex.Path.PathCommandProxy
                prev = cmd_proxy.previous_end_point
                end = cmd_proxy.end_point
                if cmd_proxy.letter == "M":
                    result.append(Move(*cmd_proxy.args))
                else:
                    for seg in self.fractalize(
                        (prev.x, prev.y, end.x, end.y),
                        self.options.subdivs,
                        self.options.smooth,
                    ):
                        result.append(Line(*seg))
                    result.append(Line(end.x, end.y))

            node.path = result

    def fractalize(self, coords, subdivs, smooth):
        """recursively subdivide the segments left and right of the subdivision"""
        subdiv_point = calculate_subdivision(smooth, *coords)

        if subdivs:
            # recursively subdivide the segment left of the subdivision point
            for left_seg in self.fractalize(
                coords[:2] + subdiv_point[-2:], subdivs - 1, smooth
            ):
                yield left_seg

            yield subdiv_point

            # recursively subdivide the segment right of the subdivision point
            for right_seg in self.fractalize(
                subdiv_point[-2:] + coords[-2:], subdivs - 1, smooth
            ):
                yield right_seg


if __name__ == "__main__":
    Fractalize().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2001-2002 Matt Chisholm matt@theory.org
# Copyright (C) 2008 Joel Holdsworth joel@airwebreathe.org.uk
#    for AP
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import cmath
import copy
import math
import os
import random
import re
import sys

import inkex
from inkex import Vector2d, load_svg

import render_alphabetsoup_config

syntax = render_alphabetsoup_config.syntax
alphabet = render_alphabetsoup_config.alphabet
units = render_alphabetsoup_config.units
font = render_alphabetsoup_config.font


def load_path(filename):
    """Loads a super-path from a given SVG file"""
    base = os.path.normpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
    # __file__ is better then sys.argv[0] because this file may be a module
    # for another one.
    fullpath = os.path.join(base, filename)
    tree = load_svg(fullpath)
    root = tree.getroot()
    elem = root.findone("svg:path")
    if elem is None:
        return None, 0, 0
    width = float(root.get("width"))
    height = float(root.get("height"))
    return (
        elem.path.to_arrays(),
        width,
        height,
    )  # Currently we only support a single path


def combinePaths(pathA, pathB):
    if pathA is None and pathB is None:
        return None
    elif pathA is None:
        return pathB
    elif pathB is None:
        return pathA
    else:
        return pathA + pathB


def reverseComponent(c):
    nc = []
    last = c.pop()
    nc.append(["M", last[1][-2:]])
    while c:
        this = c.pop()
        cmd = last[0]
        if cmd == "C":
            nc.append([last[0], last[1][2:4] + last[1][:2] + this[1][-2:]])
        else:
            nc.append([last[0], this[1][-2:]])
        last = this
    return nc


def reversePath(sp):
    rp = []
    component = []
    for p in sp:
        cmd, params = p
        if cmd == "Z":
            rp.extend(reverseComponent(component))
            rp.append(["Z", []])
            component = []
        else:
            component.append(p)
    return rp


def _lr_cb(p, width):
    p.scale(-1, 1)
    p.translate(width, 0)


def _tp_cb(p, height):
    p.scale(1, -1)
    p.translate(0, height)


def flip(sp, cb, param):
    # print('flip before +' + str(sp))
    p = inkex.Path(sp)
    cb(p, param)
    del sp[:]

    prev = Vector2d()
    prev_prev = Vector2d()
    first = Vector2d()

    for i, seg in enumerate(p):
        if i == 0:
            first = seg.end_point(first, prev)
        cps = []
        for cp in seg.control_points(first, prev, prev_prev):
            prev_prev = prev
            prev = cp
            cps.extend(cp)
        sp.append([seg.letter, cps])
    # print('flip after +' + str(sp))


def flipLeftRight(sp, width):
    return flip(sp, _lr_cb, width)


def flipTopBottom(sp, height):
    return flip(sp, _tp_cb, height)


def solveQuadratic(a, b, c):
    det = b * b - 4.0 * a * c
    if det >= 0:  # real roots
        sdet = math.sqrt(det)
    else:  # complex roots
        sdet = cmath.sqrt(det)
    return (-b + sdet) / (2 * a), (-b - sdet) / (2 * a)


def cbrt(x):
    if x >= 0:
        return x ** (1.0 / 3.0)
    else:
        return -((-x) ** (1.0 / 3.0))


def findRealRoots(a, b, c, d):
    if a != 0:
        a, b, c, d = 1, b / float(a), c / float(a), d / float(a)  # Divide through by a
        t = b / 3.0
        p, q = c - 3 * t**2, d - c * t + 2 * t**3
        u, v = solveQuadratic(1, q, -((p / 3.0) ** 3))
        if isinstance(u, complex):  # Complex Cubic Root
            r = math.sqrt(u.real**2 + u.imag**2)
            w = math.atan2(u.imag, u.real)
            y1 = 2 * cbrt(r) * math.cos(w / 3.0)
        else:  # Complex Real Root
            y1 = cbrt(u) + cbrt(v)

        y2, y3 = solveQuadratic(1, y1, p + y1**2)

        if isinstance(y2, complex):  # Are y2 and y3 complex?
            return [y1 - t]
        return [y1 - t, y2 - t, y3 - t]
    elif b != 0:
        det = c * c - 4.0 * b * d
        if det >= 0:
            return [
                (-c + math.sqrt(det)) / (2.0 * b),
                (-c - math.sqrt(det)) / (2.0 * b),
            ]
    elif c != 0:
        return [-d / c]
    return []


def mxfm(image, width, height, stack):  # returns possibly transformed image
    tbimage = image
    if stack[0] == "-":  # top-bottom flip
        flipTopBottom(tbimage, height)
        tbimage = reversePath(tbimage)
        stack.pop(0)

    lrimage = tbimage
    if stack[0] == "|":  # left-right flip
        flipLeftRight(tbimage, width)
        lrimage = reversePath(lrimage)
        stack.pop(0)
    return lrimage


def comparerule(rule, nodes):  # compare node list to nodes in rule
    for i in range(0, len(nodes)):  # range( a, b ) = (a, a+1, a+2 ... b-2, b-1)
        if nodes[i] == rule[i][0]:
            pass
        else:
            return 0
    return 1


def findrule(state, nodes):  # find the rule which generated this subtree
    ruleset = syntax[state][1]
    nodelen = len(nodes)
    for rule in ruleset:
        rulelen = len(rule)
        if (rulelen == nodelen) and (comparerule(rule, nodes)):
            return rule
    return


def generate(state):  # generate a random tree (in stack form)
    stack = [state]
    if len(syntax[state]) == 1:  # if this is a stop symbol
        return stack
    else:
        stack.append("[")
        path = random.randint(
            0, (len(syntax[state][1]) - 1)
        )  # choose randomly from next states
        for symbol in syntax[state][1][path]:  # recurse down each non-terminal
            if symbol != 0:  # 0 denotes end of list ###
                substack = generate(symbol[0])  # get subtree
                for elt in substack:
                    stack.append(elt)
                if symbol[3]:
                    stack.append("-")  # top-bottom flip
                if symbol[4]:
                    stack.append("|")  # left-right flip
            # else:
            # inkex.debug("found end of list in generate( state =", state, ")") # this should be deprecated/never happen
        stack.append("]")
        return stack


def draw(stack):  # draw a character based on a tree stack
    state = stack.pop(0)
    # print state,

    image, width, height = load_path(font + syntax[state][0])  # load the image
    if stack[0] != "[":  # terminal stack element
        if len(syntax[state]) == 1:  # this state is a terminal node
            return image, width, height
        else:
            substack = generate(state)  # generate random substack
            return draw(substack)  # draw random substack
    else:
        # inkex.debug("[")
        stack.pop(0)
        images = []  # list of daughter images
        nodes = []  # list of daughter names
        while stack[0] != "]":  # for all nodes in stack
            newstate = stack[0]  # the new state
            newimage, width, height = draw(stack)  # draw the daughter state
            if newimage:
                tfimage = mxfm(
                    newimage, width, height, stack
                )  # maybe transform daughter state
                images.append([tfimage, width, height])  # list of daughter images
                nodes.append(newstate)  # list of daughter nodes
            else:
                # inkex.debug(("recurse on",newstate,"failed")) # this should never happen
                return None, 0, 0
        rule = findrule(state, nodes)  # find the rule for this subtree

        for i in range(0, len(images)):
            currimg, width, height = images[i]

            if currimg:
                # box = inkex.Path(currimg).bounding_box()
                dx = rule[i][1] * units
                dy = rule[i][2] * units
                # newbox = ((box[0]+dx),(box[1]+dy),(box[2]+dx),(box[3]+dy))
                currimg = (inkex.Path(currimg).translate(dx, dy)).to_arrays()
                image = combinePaths(image, currimg)

        stack.pop(0)
        return image, width, height


def draw_crop_scale(stack, zoom):  # draw, crop and scale letter image
    image, width, height = draw(stack)
    bbox = inkex.Path(image).bounding_box()
    image = (inkex.Path(image).translate(-bbox.x.minimum, 0)).to_arrays()
    image = (inkex.Path(image).scale(zoom / units, zoom / units)).to_arrays()
    return image, bbox.width, bbox.height


def randomize_input_string(
    tokens, zoom
):  # generate a glyph starting from each token in the input string
    imagelist = []

    stack = None
    for i in range(0, len(tokens)):
        char = tokens[i]
        # if ( re.match("[a-zA-Z0-9?]", char)):
        if char in alphabet:
            if (i > 0) and (
                char == tokens[i - 1]
            ):  # if this letter matches previous letter
                imagelist.append(imagelist[len(stack) - 1])  # make them the same image
            else:  # generate image for letter
                stack = alphabet[char][
                    random.randint(0, (len(alphabet[char]) - 1))
                ].split(".")
                # stack = string.split( alphabet[char][random.randint(0,(len(alphabet[char])-2))] , "." )
                imagelist.append(draw_crop_scale(stack, zoom))
        elif char == " ":  # add a " " space to the image list
            imagelist.append(" ")
        else:  # this character is not in config.alphabet, skip it
            sys.stderr.write('bad character "{}"\n'.format(char))
    return imagelist


def generate_random_string(
    tokens, zoom
):  # generate a totally random glyph for each glyph in the input string
    imagelist = []
    for char in tokens:
        if char == " ":  # add a " " space to the image list
            imagelist.append(" ")
        else:
            if re.match("[a-z]", char):  # generate lowercase letter
                stack = generate("lc")
            elif re.match("[A-Z]", char):  # generate uppercase letter
                stack = generate("UC")
            else:  # this character is not in config.alphabet, skip it
                sys.stderr.write('bad character"{}"\n'.format(char))
                stack = generate("start")
            imagelist.append(draw_crop_scale(stack, zoom))

    return imagelist


def optikern(image, width, zoom):  # optical kerning algorithm
    left = []
    right = []

    resolution = 8
    for i in range(0, 18 * resolution):
        y = 1.0 / resolution * (i + 0.5) * zoom
        xmin = None
        xmax = None

        for cmd, params in image:
            if cmd == "M":
                # A move cannot contribute to the bounding box
                last = params[:]
                lastctrl = params[:]
            elif cmd == "L":
                if (last[1] <= y <= params[1]) or (params[1] <= y <= last[1]):
                    if params[0] == last[0]:
                        x = params[0]
                    else:
                        a = (params[1] - last[1]) / (params[0] - last[0])
                        b = last[1] - a * last[0]
                        if a != 0:
                            x = (y - b) / a
                        else:
                            x = None

                    if x:
                        if xmin is None or x < xmin:
                            xmin = x
                        if xmax is None or x > xmax:
                            xmax = x

                last = params[:]
                lastctrl = params[:]
            elif cmd == "C":
                if last:
                    bx0, by0 = last[:]
                    bx1, by1, bx2, by2, bx3, by3 = params[:]

                    d = by0 - y
                    c = -3 * by0 + 3 * by1
                    b = 3 * by0 - 6 * by1 + 3 * by2
                    a = -by0 + 3 * by1 - 3 * by2 + by3

                    ts = findRealRoots(a, b, c, d)

                    for t in ts:
                        if 0 <= t <= 1:
                            x = (
                                (-bx0 + 3 * bx1 - 3 * bx2 + bx3) * (t**3)
                                + (3 * bx0 - 6 * bx1 + 3 * bx2) * (t**2)
                                + (-3 * bx0 + 3 * bx1) * t
                                + bx0
                            )
                            if xmin is None or x < xmin:
                                xmin = x
                            if xmax is None or x > xmax:
                                xmax = x

                last = params[-2:]
                lastctrl = params[2:4]

            elif cmd == "Q":
                # Quadratic beziers are ignored
                last = params[-2:]
                lastctrl = params[2:4]

            elif cmd == "A":
                # Arcs are ignored
                last = params[-2:]
                lastctrl = params[2:4]

        if xmin is not None and xmax is not None:
            left.append(xmin)  # distance from left edge of region to left edge of bbox
            right.append(
                width - xmax
            )  # distance from right edge of region to right edge of bbox
        else:
            left.append(width)
            right.append(width)

    return left, right


def layoutstring(
    imagelist, zoom
):  # layout string of letter-images using optical kerning
    kernlist = []
    length = zoom
    for entry in imagelist:
        if entry == " ":  # leaving room for " " space characters
            length = length + (zoom * render_alphabetsoup_config.space)
        else:
            image, width, height = entry
            length = length + width + zoom  # add letter length to overall length
            kernlist.append(
                optikern(image, width, zoom)
            )  # append kerning data for this image

    workspace = None

    position = zoom
    for i in range(0, len(kernlist)):
        while imagelist[i] == " ":
            position = position + (zoom * render_alphabetsoup_config.space)
            imagelist.pop(i)
        image, width, height = imagelist[i]

        # set the kerning
        if i == 0:
            kern = 0  # for first image, kerning is zero
        else:
            kerncompare = []  # kerning comparison array
            for j in range(0, len(kernlist[i][0])):
                kerncompare.append(kernlist[i][0][j] + kernlist[i - 1][1][j])
            kern = min(kerncompare)

        position = position - kern  # move position back by kern amount
        thisimage = copy.deepcopy(image)
        thisimage = (inkex.Path(thisimage).translate(position, 0)).to_arrays()
        workspace = combinePaths(workspace, thisimage)
        position = position + width + zoom  # advance position by letter width

    return workspace


def tokenize(text):
    """Tokenize the string, looking for LaTeX style, multi-character tokens in the string, like \\yogh."""
    tokens = []
    i = 0
    while i < len(text):
        c = text[i]
        i += 1
        if c == "\\":  # found the beginning of an escape
            t = ""
            while i < len(text):  # gobble up content of the escape
                c = text[i]
                if c == "\\":  # found another escape, stop this one
                    break
                i += 1
                if c == " ":  # a space terminates this escape
                    break
                t += c  # stick this character onto the token
            if t:
                tokens.append(t)
        else:
            tokens.append(c)
    return tokens


class AlphabetSoup(inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument(
            "-t", "--text", default="Inkscape", help="The text for alphabet soup"
        )
        pars.add_argument(
            "-z", "--zoom", type=float, default=8.0, help="The zoom on the output"
        )
        pars.add_argument(
            "-r",
            "--randomize",
            type=inkex.Boolean,
            default=False,
            help="Generate random (unreadable) text",
        )

    def effect(self):
        zoom = self.svg.unittouu(str(self.options.zoom) + "px")

        if self.options.randomize:
            imagelist = generate_random_string(self.options.text, zoom)
        else:
            tokens = tokenize(self.options.text)
            imagelist = randomize_input_string(tokens, zoom)

        image = layoutstring(imagelist, zoom)

        if image:
            s = {"stroke": "none", "fill": "#000000"}

            new = inkex.PathElement(style=str(inkex.Style(s)), d=str(inkex.Path(image)))

            layer = self.svg.get_current_layer()
            layer.append(new)

            # compensate preserved transforms of parent layer
            if layer.getparent() is not None:
                mat = (
                    self.svg.get_current_layer().transform
                    @ inkex.Transform([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
                ).matrix
                new.transform @= -inkex.Transform(mat)


if __name__ == "__main__":
    AlphabetSoup().run()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../"))

import datetime

# -- Project information -----------------------------------------------------

project = "inkex"
copyright = f"{datetime.datetime.now().year} The Inkscape Project"
author = "The Inkscape Project"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinx.ext.napoleon",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "show_toc_level": 2,
    "use_edit_page_button": True,
    "gitlab_url": "https://gitlab.com/inkscape/extensions",
}

html_context = {
    "gitlab_user": "inkscape",
    "gitlab_repo": "extensions",
    "gitlab_version": "master",
    "doc_path": "docs",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# -- Extension configuration -------------------------------------------------

always_document_param_types = True

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

autodoc_member_order = "bysource"

autodoc_preserve_defaults = True

html_favicon = "favicon.svg"



# coding=utf-8
#
# Copyright (C) 2018 - Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Deprecation functionality which does not require imports from Inkex.
"""

import os
import traceback
import warnings
from typing import Optional

try:
    DEPRECATION_LEVEL = int(os.environ.get("INKEX_DEPRECATION_LEVEL", 1))
except ValueError:
    DEPRECATION_LEVEL = 1


def _deprecated(msg, stack=2, level=DEPRECATION_LEVEL):
    """Internal method for raising a deprecation warning"""
    if level > 1:
        msg += " ; ".join(traceback.format_stack())
    if level:
        warnings.warn(msg, category=DeprecationWarning, stacklevel=stack + 1)


def deprecate(func, version: Optional[str] = None):
    r"""Function decorator for deprecation functions which have a one-liner
    equivalent in the new API. The one-liner has to passed as a string
    to the decorator.

    >>> @deprecate
    >>> def someOldFunction(*args):
    >>>     '''Example replacement code someNewFunction('foo', ...)'''
    >>>     someNewFunction('foo', *args)

    Or if the args API is the same:

    >>> someOldFunction = deprecate(someNewFunction)

    """

    def _inner(*args, **kwargs):
        _deprecated(f"{func.__module__}.{func.__name__} -> {func.__doc__}", stack=2)
        return func(*args, **kwargs)

    _inner.__name__ = func.__name__
    if func.__doc__:
        if version is None:
            _inner.__doc__ = "Deprecated -> " + func.__doc__
        else:
            _inner.__doc__ = f"""{func.__doc__}\n\n.. deprecated:: {version}\n"""
    return _inner


class DeprecatedSvgMixin:
    """Mixin which adds deprecated API elements to the SvgDocumentElement"""

    @property
    def selected(self):
        """svg.selection"""
        return self.selection

    @deprecate
    def set_selected(self, *ids):
        r"""svg.selection.set(\*ids)"""
        return self.selection.set(*ids)

    @deprecate
    def get_z_selected(self):
        """svg.selection.rendering_order()"""
        return self.selection.rendering_order()

    @deprecate
    def get_selected(self, *types):
        r"""svg.selection.filter(\*types).values()"""
        return self.selection.filter(*types).values()

    @deprecate
    def get_selected_or_all(self, *types):
        """Set select_all = True in extension class"""
        if not self.selection:
            self.selection.set_all()
        return self.selection.filter(*types)

    @deprecate
    def get_selected_bbox(self):
        """selection.bounding_box()"""
        return self.selection.bounding_box()

    @deprecate
    def get_first_selected(self, *types):
        r"""selection.filter(\*types).first() or [0] if you'd like an error"""
        return self.selection.filter(*types).first()



#!/usr/bin/env python3
#
# Copyright (C) 2007-2011 Rob Antonishen; rob.antonishen@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import random

import inkex
from inkex.utils import KeyDict
from inkex import SvgDocumentElement
from inkex.localization import inkex_gettext as _

# Old settings, supported because users click 'ok' without looking.
XAN = KeyDict({"l": "left", "r": "right", "m": "center_x"})
YAN = KeyDict({"t": "top", "b": "bottom", "m": "center_y"})
CUSTOM_DIRECTION = {270: "tb", 90: "bt", 0: "lr", 360: "lr", 180: "rl"}


class Restack(inkex.EffectExtension):
    """Change the z-order of objects based on their position on the canvas"""

    restack_help = staticmethod(lambda: None)

    def add_arguments(self, pars):
        pars.add_argument(
            "--tab", type=self.arg_method("restack"), default=self.restack_positional
        )
        pars.add_argument("--direction", default="lr", help="direction to restack")
        pars.add_argument("--angle", type=float, default=0.0, help="arbitrary angle")
        pars.add_argument("--xanchor", default="l", help="horizontal point to compare")
        pars.add_argument("--yanchor", default="t", help="vertical point to compare")
        pars.add_argument(
            "--zsort", default="rev", help="Restack mode based on Z-Order"
        )
        pars.add_argument("--nb_direction", default="", help="Direction tab")

    def effect(self):
        if not self.svg.selection:
            raise inkex.AbortExtension(_("There is no selection to restack."))

        # process selection to get list of objects to be arranged
        parentnode = None
        for node in self.svg.selection.filter(SvgDocumentElement):
            parentnode = node
            self.svg.set_selection(*list(node))

        if parentnode is None:
            parentnode = self.svg.get_current_layer()

        self.options.tab(parentnode)

    def restack_positional(self, parentnode):
        """Restack based on canvas position"""
        # move them to the top of the object stack in this order.
        for node in sorted(self.svg.selection.values(), key=self._sort):
            parentnode.append(node)
        return True

    def _sort(self, node):
        x, y = self.options.xanchor, self.options.yanchor
        selbox = self.svg.selection.bounding_box()
        direction = self.options.direction
        if "custom" in self.options.nb_direction:
            direction = self.options.angle
        return node.bounding_box().get_anchor(x, y, direction, selbox)

    def restack_z_order(self, parentnode):
        """Restack based on z-order"""
        objects = list(self.svg.selection.rendering_order())
        if self.options.zsort == "rev":
            objects.reverse()
        elif self.options.zsort == "rand":
            random.shuffle(objects)
        if parentnode is not None:
            for item in objects:
                parentnode.append(item)
        return True


if __name__ == "__main__":
    Restack().run()



#
# Copyright 2015 Ian Denhardt <ian@zenhack.net>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Convenience library for concurrency

GUI apps frequently need concurrency, for example to avoid blocking UI while
doing some long running computation. This module provides helpers for doing
this kind of thing.

The functions/methods here which spawn callables asynchronously
don't supply a direct way to provide arguments. Instead, the user is
expected to use a lambda, e.g::

    holding(lck, lambda: do_stuff(1,2,3, x='hello'))

This is because the calling function may have additional arguments which
could obscure the user's ability to pass arguments expected by the called
function. For example, in the call::

    holding(lck, lambda: run_task(blocking=True), blocking=False)

the blocking argument to holding might otherwise conflict with the
blocking argument to run_task.
"""

import time
import threading
from datetime import datetime, timedelta

from functools import wraps
from typing import Any, Tuple
from gi.repository import Gdk, GLib


class Future:
    """A deferred result

    A `Future` is a result-to-be; it can be used to deliver a result
    asynchronously. Typical usage:

        >>> def background_task(task):
        ...     ret = Future()
        ...     def _task(x):
        ...         return x - 4 + 2
        ...     thread = threading.Thread(target=lambda: ret.run(lambda: _task(7)))
        ...     thread.start()
        ...     return ret
        >>> # Do other stuff
        >>> print(ret.wait())
        5

    :func:`run` will also propagate exceptions; see its docstring for details.
    """

    def __init__(self):
        self._lock = threading.Lock()
        self._value = None
        self._exception = None
        self._lock.acquire()

    def is_ready(self):
        """Return whether the result is ready"""
        result = self._lock.acquire(False)
        if result:
            self._lock.release()
        return result

    def wait(self):
        """Wait for the result.

        `wait` blocks until the result is ready (either :func:`result` or
        :func:`exception` has been called), and then returns it (in the case
        of :func:`result`), or raises it (in the case of :func:`exception`).
        """
        with self._lock:
            if self._exception is None:
                return self._value
            else:
                raise self._exception  # pylint: disable=raising-bad-type

    def result(self, value):
        """Supply the result as a return value.

        ``value`` is the result to supply; it will be returned when
        :func:`wait` is called.
        """
        self._value = value
        self._lock.release()

    def exception(self, err):
        """Supply an exception as the result.

        Args:
            err (Exception): an exception, which will be raised when :func:`wait`
                is called.
        """
        self._exception = err
        self._lock.release()

    def run(self, task):
        """Calls task(), and supplies the result.

        If ``task`` raises an exception, pass it to :func:`exception`.
        Otherwise, pass the return value to :func:`result`.
        """
        try:
            self.result(task())
        except Exception as err:  # pylint: disable=broad-except
            self.exception(err)


class DebouncedSyncVar:
    """A synchronized variable, which debounces its value

    :class:`DebouncedSyncVar` supports three operations: put, replace, and get.
    get will only retrieve a value once it has "settled," i.e. at least
    a certain amount of time has passed since the last time the value
    was modified.
    """

    def __init__(self, delay_seconds=0):
        """Create a new dsv with the supplied delay, and no initial value."""
        self._cv = threading.Condition()
        self._delay = timedelta(seconds=delay_seconds)

        self._deadline = None
        self._value = None

        self._have_value = False

    def set_delay(self, delay_seconds):
        """Set the delay in seconds of the debounce."""
        with self._cv:
            self._delay = timedelta(seconds=delay_seconds)

    def get(self, blocking=True, remove=True) -> Tuple[Any, bool]:
        """Retrieve a value.

        Args:
            blocking (bool, optional): if True, block until (1) the dsv has a value
                and (2) the value has been unchanged for an amount of time greater
                than or equal to the dsv's delay. Otherwise, if these conditions
                are not met, return ``(None, False)`` immediately. Defaults to True.
            remove (bool, optional): if True, remove the value when returning it.
                Otherwise, leave it where it is.. Defaults to True.

        Returns:
            Tuple[Any, bool]: Tuple (value, ok). ``value`` is the value of the variable
            (if successful, see above), and ok indicates whether or not a value was
            successfully retrieved.
        """
        while True:
            with self._cv:
                # If there's no value, either wait for one or return
                # failure.
                while not self._have_value:
                    if blocking:
                        self._cv.wait()
                    else:
                        return None, False  # pragma: no cover

                now = datetime.now()
                deadline = self._deadline
                value = self._value
                if deadline <= now:
                    # Okay, we're good. Remove the value if necessary, and
                    # return it.
                    if remove:
                        self._have_value = False
                        self._value = None
                    self._cv.notify()
                    return value, True

            # Deadline hasn't passed yet. Either wait or return failure.
            if blocking:
                time.sleep((deadline - now).total_seconds())
            else:
                return None, False  # pragma: no cover

    def replace(self, value):
        """Replace the current value of the dsv (if any) with ``value``.

        replace never blocks (except briefly to acquire the lock). It does not
        wait for any unit of time to pass (though it does reset the timer on
        completion), nor does it wait for the dsv's value to appear or
        disappear.
        """
        with self._cv:
            self._replace(value)

    def put(self, value):
        """Set the dsv's value to ``value``.

        If the dsv already has a value, this blocks until the value is removed.
        Upon completion, this resets the timer.
        """
        with self._cv:
            while self._have_value:
                self._cv.wait()
            self._replace(value)

    def _replace(self, value):
        self._have_value = True
        self._value = value
        self._deadline = datetime.now() + self._delay
        self._cv.notify()


def spawn_thread(func):
    """Call ``func()`` in a separate thread

    Returns the corresponding :class:`threading.Thread` object.
    """
    thread = threading.Thread(target=func)
    thread.start()
    return thread


def in_mainloop(func):
    """Run f() in the gtk main loop

    Returns a :class:`Future` object which can be used to retrieve the return
    value of the function call.

    :func:`in_mainloop` exists because Gtk isn't threadsafe, and therefore cannot be
    manipulated except in the thread running the Gtk main loop. :func:`in_mainloop`
    can be used by other threads to manipulate Gtk safely.
    """
    future = Future()

    def handler(*_args, **_kwargs):
        """Function to be called in the future"""
        future.run(func)

    Gdk.threads_add_idle(0, handler, None)
    return future


def mainloop_only(f):
    """A decorator which forces a function to only be run in Gtk's main loop.

    Invoking a decorated function as ``f(*args, **kwargs)`` is equivalent to
    using the undecorated function (from a thread other than the one running
    the Gtk main loop) as::

        in_mainloop(lambda: f(*args, **kwargs)).wait()

    :func:`mainloop_only` should be used to decorate functions which are unsafe
    to run outside of the Gtk main loop.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if GLib.main_depth():
            # Already in a mainloop, so just run it.
            return f(*args, **kwargs)
        return in_mainloop(lambda: f(*args, **kwargs)).wait()

    return wrapper


def holding(lock, task, blocking=True):
    """Run task() while holding ``lock``.

    Args:
        blocking (bool, optional): if True, wait for the lock before running.
            Otherwise, if the lock is busy, return None immediately, and don't
            spawn `task`. Defaults to True.

    Returns:
        Union[Future, None]: The return value is a future which can be used to retrieve
        the result of running task (or None if the task was not run).
    """
    if not lock.acquire(False):
        return None
    ret = Future()

    def _target():
        ret.run(task)
        if ret._exception:  # pragma: no cover
            ret.wait()
        lock.release()

    threading.Thread(target=_target).start()
    return ret


def run_or_wait(func):
    """A decorator which runs the function using :func:`holding`

    This function creates a single lock for this function and
    waits for the lock to release before returning.

    See :func:`holding` above, with ``blocking=True``
    """
    lock = threading.Lock()

    def _inner(*args, **kwargs):
        return holding(lock, lambda: func(*args, **kwargs), blocking=True)

    return _inner


def run_or_none(func):
    """A decorator which runs the function using :func:`holding`

    This function creates a single lock for this function and
    returns None if the process is already running (locked)

    See :func:`holding` above with ``blocking=True``
    """
    lock = threading.Lock()

    def _inner(*args, **kwargs):
        return holding(lock, lambda: func(*args, **kwargs), blocking=False)

    return _inner



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2013 Nicolas Dufour (jazzynico)
# Direction code from the Restack extension, by Rob Antonishen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""
Merge text blocks together.
"""

import inkex
from inkex.utils import KeyDict
from inkex import Rectangle, FlowRoot, FlowPara, FlowRegion, TextElement, Tspan

# Old settings, supported because users click 'ok' without looking.
XAN = KeyDict({"l": "left", "r": "right", "m": "center_x"})
YAN = KeyDict({"t": "top", "b": "bottom", "m": "center_y"})


class Merge(inkex.EffectExtension):
    """Merge text blocks together"""

    def add_arguments(self, pars):
        pars.add_argument(
            "-d", "--direction", default="lr", help="direction to merge text"
        )
        pars.add_argument(
            "-x", "--xanchor", default="left", help="horiz point to compare"
        )
        pars.add_argument(
            "-y", "--yanchor", default="top", help="vertical point to compare"
        )
        pars.add_argument("-k", "--keepstyle", type=inkex.Boolean, help="keep format")
        pars.add_argument(
            "-t",
            "--flowtext",
            type=inkex.Boolean,
            help="use a flow text structure instead of a normal text element",
        )

    def effect(self):
        if not self.svg.selection:
            for node in self.svg.xpath("//svg:text | //svg:flowRoot"):
                self.svg.selection[node.get("id")] = node

        if not self.svg.selection:
            return

        parentnode = self.svg.get_current_layer()

        if self.options.flowtext:
            text_element = FlowRoot
            text_span = FlowPara
        else:
            text_element = TextElement
            text_span = Tspan

        text_root = parentnode.add(text_element())
        text_root.set("xml:space", "preserve")
        text_root.style = {
            "font-size": "20px",
            "font-style": "normal",
            "font-weight": "normal",
            "line-height": "125%",
            "letter-spacing": "0px",
            "word-spacing": "0px",
            "fill": "#000000",
            "fill-opacity": 1,
            "stroke": "none",
        }

        for node in sorted(self.svg.selection.values(), key=self._sort):
            self.recurse(text_span, node, text_root)

        if self.options.flowtext:
            region = text_root.add(FlowRegion())
            region.set("xml:space", "preserve")
            rect = region.add(Rectangle())
            rect.set("xml:space", "preserve")
            rect.set("height", 200)
            rect.set("width", 200)

    def _sort(self, node):
        return node.bounding_box().get_anchor(
            self.options.xanchor, self.options.yanchor, self.options.direction
        )

    def recurse(self, text_span, node, span):
        """Recursively go through each node self calling on child nodes"""
        if not isinstance(node, FlowRegion):
            newspan = span.add(text_span())
            newspan.set("xml:space", "preserve")

            newspan.set("sodipodi:role", node.get("sodipodi:role"))
            if isinstance(node, (TextElement, FlowPara)):
                newspan.set("sodipodi:role", "line")

            if self.options.keepstyle:
                newspan.style = node.style

            if node.text is not None:
                newspan.text = node.text
            for child in node:
                self.recurse(text_span, child, newspan)
            if node.tail and not isinstance(node, TextElement):
                newspan.tail = node.tail


if __name__ == "__main__":
    Merge().run()



#!/usr/bin/env python3
#
# Copyright 2008, 2009 Hannes Hochreiner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.
#
"""Automatic text for jessyInk"""

import inkex
from jessyink_install import JessyInkMixin, _


class AutoTexts(JessyInkMixin, inkex.EffectExtension):
    """Add AutoText to jessyInk"""

    def add_arguments(self, pars):
        pars.add_argument("--tab", dest="what")
        pars.add_argument("--autoText", default="none")

    def effect(self):
        self.is_installed()

        if not self.svg.selection:
            inkex.errormsg(_("To assign an effect, please select an object.\n\n"))

        for node in self.svg.selection.get(inkex.Tspan):
            if self.options.autoText == "slideTitle":
                node.set("jessyink:autoText", "slideTitle")
            elif self.options.autoText == "slideNumber":
                node.set("jessyink:autoText", "slideNumber")
            elif self.options.autoText == "numberOfSlides":
                node.set("jessyink:autoText", "numberOfSlides")
            elif node.get("jessyink:autoText"):
                node.set("jessyink:autoText", None)


if __name__ == "__main__":
    AutoTexts().run()
